import io
import json
import ssl
import traceback
import uuid
from typing import Any, Dict, List, Optional

import aiohttp
import pyalex
from pyalex import Authors, Institutions, Sources, Works
from pypdf import PdfReader
from .papers import get_paper_details

from .models import (
    ENTITY_TYPES,
    INSTITUTION_TYPES,
    PUBLICATION_TYPES,
    SOURCE_TYPES,
    WORK_TYPES,
    Author,
    Entity,
    Institution,
    Publication,
    Publisher,
    SearchInput,
    Work,
)

from .http import ThrottledClientSession, validate_url
from .scrape import browser_scrape
from .google import search_google
from .user_agent import get_user_agent_header

import logging

# Set up logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

# Configure SSL context
ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE


def _parse_openalex_id(openalex_id: str) -> str:
    """
    Parse OpenAlex ID and return a string representation of UUID.

    Parameters
    ----------
    openalex_id : str
        The OpenAlex ID to parse.

    Returns
    -------
    str
        A string representation of the UUID generated from the OpenAlex ID.

    Notes
    -----
    This function extracts the last part of the OpenAlex ID and uses it to generate
    a UUID5 using the URL namespace.
    """
    id_part = openalex_id.split("/")[-1]
    return str(uuid.uuid5(uuid.NAMESPACE_URL, id_part))

def try_get_url(entity: Dict[str, Any]) -> Optional[str]:
    """
    Try to get the URL from the entity.

    Parameters
    ----------
    entity : Dict[str, Any]
        The dictionary containing the entity data from OpenAlex.
    """
    url = (
        entity.get("open_access", {}).get("oa_url")
        or entity.get("")
        or entity.get("primary_location", {}).get("pdf_url")
        or entity.get("doi")
        or entity.get("url")
        or entity.get("ids", {}).get("doi")
        or entity.get("ids", {}).get("pmid")
        or entity.get("ids", {}).get("arxiv")
        or entity.get("ids", {}).get("openalex")
    )
    return validate_url(url) if url else None



async def _process_work(entity: Dict[str, Any]) -> Optional[Work]:
    url = try_get_url(entity)

    if "arxiv.org" in url or "biorxiv.org" in url or "pubmed.ncbi.nlm.nih.gov" in url:
        paper_details = await get_paper_details(url)
        if paper_details:
            return Work(
                id=_parse_openalex_id(entity["id"]),
                work_type=paper_details.work_type,
                source_type=paper_details.source_type,
                name=paper_details.name,
                url=paper_details.url,
                abstract=paper_details.abstract,
                full_text=paper_details.full_text,
                year=paper_details.year,
                doi=paper_details.doi,
                authors=paper_details.authors,
                institutions=paper_details.institutions,
                publications=paper_details.publications,
            )

    authors = []
    all_institutions = []

    for author_data in entity.get("authorships", []):
        author = Author(
            id=_parse_openalex_id(author_data["author"].get("id", "")),
            name=author_data["author"].get("display_name", ""),
            source_type=SOURCE_TYPES.OPENALEX,
        )
        author_institutions = [
            Institution(
                id=_parse_openalex_id(inst.get("id", "")),
                name=inst.get("display_name", ""),
                source_type=SOURCE_TYPES.OPENALEX,
                institution_type=INSTITUTION_TYPES.ACADEMIC,
            )
            for inst in author_data.get("institutions", [])
        ]
        author.institutions = author_institutions
        authors.append(author)

        all_institutions.extend(author_institutions)

    work_institutions = [
        Institution(
            id=_parse_openalex_id(inst.get("id", "")),
            name=inst.get("display_name", ""),
            source_type=SOURCE_TYPES.OPENALEX,
            institution_type=INSTITUTION_TYPES.ACADEMIC,
        )
        for inst in entity.get("institutions", [])
    ]

    all_institutions.extend(work_institutions)

    publications = []
    locations = entity.get("locations", [])
    if entity.get("primary_location"):
        locations.insert(0, entity["primary_location"])

    for location in locations:
        if location.get("source"):
            source = location["source"]
            publication_type = _get_publication_type(source.get("type"))
            
            if publication_type == PUBLICATION_TYPES.CONFERENCE_PROCEEDINGS:
                conference = Institution(
                    id=_parse_openalex_id(source.get("id", "")),
                    name=source.get("display_name", ""),
                    source_type=SOURCE_TYPES.OPENALEX,
                    institution_type=INSTITUTION_TYPES.CONFERENCE,
                    url=source.get("homepage_url", None)
                )
                all_institutions.append(conference)
            
            publication = Publication(
                id=_parse_openalex_id(source.get("id", "")),
                name=source.get("display_name", ""),
                source_type=SOURCE_TYPES.OPENALEX,
                publication_type=publication_type,
                url=source.get("homepage_url", ""),
            )
            
            if source.get("host_organization"):
                host_org = source["host_organization"]
                publication.publisher = Publisher(
                    id=_parse_openalex_id(host_org),
                    name=host_org,
                    source_type=SOURCE_TYPES.OPENALEX,
                    url=None,
                    description=None,
                )
            publications.append(publication)

    work_type = WORK_TYPES.PAPER

    try:
        return Work(
            id=_parse_openalex_id(entity["id"]),
            work_type=work_type,
            source_type=SOURCE_TYPES.OPENALEX,
            name=entity.get("title", ""),
            url=validate_url(url),
            abstract=_reconstruct_abstract(entity.get("abstract_inverted_index")),
            full_text="",
            year=entity.get("publication_year"),
            doi=entity.get("doi"),
            authors=authors,
            institutions=all_institutions,
            publications=publications,
        )
    except Exception as e:
        print("Error processing work for ", entity["id"], ": ", e)
        print("Entity is: ", entity)
        return None

def _get_publication_type(type_str: str) -> PUBLICATION_TYPES:
    """
    Map OpenAlex publication types to our PUBLICATION_TYPES enum.

    Parameters
    ----------
    type_str : str
        The publication type string from OpenAlex.

    Returns
    -------
    PUBLICATION_TYPES
        The corresponding PUBLICATION_TYPES enum value.
    """
    type_mapping = {
        "journal": PUBLICATION_TYPES.JOURNAL,
        "repository": PUBLICATION_TYPES.OTHER,
        "conference": PUBLICATION_TYPES.CONFERENCE_PROCEEDINGS,
        "book": PUBLICATION_TYPES.BOOK,
    }
    return type_mapping.get(type_str, PUBLICATION_TYPES.UNKNOWN)


def _process_author(entity: Dict[str, Any]) -> Author:
    """
    Process an author entity from OpenAlex and convert it to an Author object.

    Parameters
    ----------
    entity : Dict[str, Any]
        The dictionary containing the author entity data from OpenAlex.

    Returns
    -------
    Author
        An Author object constructed from the OpenAlex data.

    Notes
    -----
    This function extracts the author's name and associated institutions from
    the OpenAlex data and creates an Author object.
    """
    return Author(
        id=_parse_openalex_id(entity["id"]),
        source_type=SOURCE_TYPES.OPENALEX,
        name=entity.get("display_name"),
        institutions=[
            _process_institution(inst) for inst in entity.get("institutions", [])
        ],
    )


def _process_institution(openalex_entity: Dict[str, Any]) -> Institution:
    """
    Process an institution entity from OpenAlex and convert it to an Institution object.

    Parameters
    ----------
    openalex_entity : Dict[str, Any]
        The dictionary containing the institution entity data from OpenAlex.

    Returns
    -------
    Institution
        An Institution object constructed from the OpenAlex data.

    Notes
    -----
    This function extracts the institution's name and ROR (Research Organization Registry)
    URL from the OpenAlex data and creates an Institution object.
    """
    return Institution(
        id=_parse_openalex_id(openalex_entity["id"]),
        source_type=SOURCE_TYPES.OPENALEX,
        institution_type=INSTITUTION_TYPES.ACADEMIC,
        name=openalex_entity.get("display_name"),
        url=openalex_entity.get("ror"),
    )


def _process_publication(entity: Dict[str, Any]) -> Publication:
    """
    Process a publication entity from OpenAlex and convert it to a Publication object.

    Parameters
    ----------
    entity : Dict[str, Any]
        The dictionary containing the publication entity data from OpenAlex.

    Returns
    -------
    Publication
        A Publication object constructed from the OpenAlex data.

    Notes
    -----
    This function extracts the publication's name and URL (homepage or general URL)
    from the OpenAlex data and creates a Publication object.
    """
    url = try_get_url(entity)
    return Publication(
        id=_parse_openalex_id(entity["id"]),
        source_type=SOURCE_TYPES.OPENALEX,
        publication_type=PUBLICATION_TYPES.JOURNAL,
        name=entity.get("display_name"),
        url=validate_url(url) if url else None,
    )


async def _process_entity(entity: Dict[str, Any], entity_type: str) -> Optional[Entity]:
    """
    Process an entity from OpenAlex based on its type.

    Parameters
    ----------
    entity : Dict[str, Any]
        The dictionary containing the entity data from OpenAlex.
    entity_type : str
        The type of the entity ('work', 'author', 'institution', or 'publication').

    Returns
    -------
    Optional[Entity]
        An Entity object (Work, Author, Institution, or Publication) if processing
        is successful, None otherwise.

    Notes
    -----
    This function serves as a router to call the appropriate processing function
    based on the entity type. It handles unknown entity types by logging a warning
    and returning None.
    """
    if entity_type == "work":
        return await _process_work(entity)
    elif entity_type == "author":
        return _process_author(entity)
    elif entity_type == "institution":
        return _process_institution(entity)
    elif entity_type == "publication":
        return _process_publication(entity)
    else:
        logging.error(f"Error processing {entity_type} entity: {str(entity.type)}")
        logging.error(traceback.format_exc())
        return None


async def _download_and_extract_text(url: str, session: aiohttp.ClientSession) -> str:
    """
    Download content from a URL and extract text, handling both PDFs and web pages.

    Parameters
    ----------
    url : str
        The URL to download content from.
    session : aiohttp.ClientSession
        The aiohttp client session to use for making requests.

    Returns
    -------
    str
        The extracted text content from the URL.

    Raises
    ------
    ValueError
        If the input URL is not a string.
    Exception
        If the download fails or content extraction encounters an error.

    Notes
    -----
    This function attempts to download the content from the given URL. If the content
    is a PDF, it extracts the text using pypdf. For other content types, it uses
    browser_scrape to extract the text.
    """
    if not isinstance(url, str):
        raise ValueError(f"Invalid URL type: {type(url)}. Expected str.")

    async with session.get(url, allow_redirects=True) as response:
        if response.status == 200:
            content_type = response.headers.get("Content-Type", "").lower()
            if "application/pdf" in content_type:
                pdf_content = await response.read()
                pdf_file = io.BytesIO(pdf_content)
                try:
                    pdf_reader = PdfReader(pdf_file)
                    text = "\n".join(page.extract_text() for page in pdf_reader.pages)
                    return text
                except Exception as e:
                    logging.error(f"Error extracting text from PDF: {e}")
            else:
                return await browser_scrape(url)
        else:
            raise Exception(
                f"Failed to download content from {url}. Status code: {response.status}"
            )


def _reconstruct_abstract(
    inverted_index: Optional[Dict[str, List[int]]]
) -> Optional[str]:
    """
    Reconstruct the abstract from an inverted index.

    Parameters
    ----------
    inverted_index : Optional[Dict[str, List[int]]]
        A dictionary where keys are words and values are lists of positions.

    Returns
    -------
    Optional[str]
        The reconstructed abstract as a string, or None if the input is None.

    Notes
    -----
    This function takes an inverted index (where words are keys and their positions
    in the text are values) and reconstructs the original text by placing each word
    in its correct position.
    """
    if not inverted_index:
        return None

    words = []
    max_position = max(max(positions) for positions in inverted_index.values())

    for i in range(max_position + 1):
        for word, positions in inverted_index.items():
            if i in positions:
                words.append(word)
                break

    return " ".join(words)


async def download_from_openalex(work: Work, session: ThrottledClientSession) -> Work:
    """
    Download the full text content for a given work from OpenAlex.

    Parameters
    ----------
    work : Work
        The Work object for which to download the full text.
    session : ThrottledClientSession
        The aiohttp client session to use for making requests.

    Returns
    -------
    Work
        The Work object with the full text content if successfully downloaded, or the original Work object if not.

    Notes
    -----
    This function attempts to download the full text using several methods:
    1. Directly from the work's URL
    2. By searching Google for a PDF of the work
    3. By scraping the work's webpage
    If all methods fail, it returns the original Work object.
    """
    print("Downloading from OpenAlex: ", work)
    if not work.url:
        return work

    try:
        # Attempt to download and extract text from the primary URL
        full_text = await _download_and_extract_text(work.url, session)
        if full_text:
            work.full_text = full_text
            return work
    except Exception as e:
        logging.warning(f"Warning: Could not download full text from primary URL: {e}")

    # If primary URL fails, search Google for a PDF
    search_results = await search_google(
        SearchInput(
            query=work.name,
            file_type="pdf",
            num_results=3,
            source_type=SOURCE_TYPES.OPENALEX,
            entity_types=[ENTITY_TYPES.WORK],
        )
    )
    for search_result in search_results:
        try:
            text = await _download_and_extract_text(search_result.url, session)
            if text:
                work.url = validate_url(search_result.url)
                logging.info(
                    f"Downloaded full text from Google search: {search_result.url}"
                )
                work.full_text = text
                return work
        except Exception as e:
            logging.warn(f"Error downloading PDF from Google search: {e}")

    try:
        # If all else fails, attempt to scrape the text from the webpage
        text = await browser_scrape(str(work.url))
        if text:
            logging.info(f"Downloaded full text from original URL: {work.url}")
            work.full_text = text
            return work
    except Exception as e:
        logging.error(f"Error scraping full text: {e}")

    return work


async def search_and_download_from_openalex(search_input: SearchInput) -> List[Entity]:
    """
    Perform a search on OpenAlex and download full text content for the results.

    Parameters
    ----------
    search_input : SearchInput
        An object containing search parameters such as query, number of results,
        and entity types to search for.

    Returns
    -------
    List[Entity]
        A list of Entity objects (Works, Authors, Institutions, Publications)
        found and processed from the OpenAlex search.

    Notes
    -----
    This function performs a search using the OpenAlex API, processes the results
    into Entity objects, and attempts to download full text content for Work entities.
    It also handles deduplication of publications and ensures all related entities
    (authors, institutions) are included in the results.
    """
    logging.info(f"Starting search with input: {search_input}")
    search_results = await search_openalex(search_input)

    entities: List[Entity] = []
    publications_set = set()  # To keep track of unique publications

    # Process search results and collect all related entities
    for result in search_results:
        logging.info(f"Processing result of type: {result.type}")
        entities.append(result)

        if isinstance(result, Work):
            for author in result.authors:
                if author not in entities:
                    entities.append(author)
                for institution in author.institutions:
                    if institution not in entities:
                        entities.append(institution)

            for institution in result.institutions:
                if institution not in entities:
                    entities.append(institution)

            for publication in result.publications:
                if publication.id not in publications_set:
                    entities.append(publication)
                    publications_set.add(publication.id)

    # Download full text for Work entities
    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=get_user_agent_header()
    ) as session:
        for entity in entities:
            if isinstance(entity, Work):
                try:
                    entity = await download_from_openalex(entity, session)
                except Exception as e:
                    logging.warn(f"Error downloading full text for {entity.name}: {e}")
                    entity.full_text = ""

    logging.info(f"Total entities created: {len(entities)}")
    return entities


async def search_openalex(search_input: SearchInput) -> List[Entity]:
    """
    Perform a search on OpenAlex for various entity types.

    Parameters
    ----------
    search_input : SearchInput
        An object containing search parameters such as query, number of results,
        and entity types to search for.

    Returns
    -------
    List[Entity]
        A list of Entity objects found in the OpenAlex search.

    Notes
    -----
    This function searches OpenAlex for different entity types (work, author,
    institution, publication) based on the search input. It uses the pyalex
    library to interact with the OpenAlex API and processes the results into
    appropriate Entity objects.
    """
    query = search_input.query
    num_results = search_input.num_results
    entity_types = search_input.entity_types

    results = []

    for entity_type in entity_types:
        try:
            logging.info(f"Searching for {entity_type} with query: {query}")
            if entity_type == "work":
                entities = Works().search(query).paginate(per_page=num_results * 2)
            elif entity_type == "author":
                entities = Authors().search(query).paginate(per_page=num_results)
            elif entity_type == "institution":
                entities = Institutions().search(query).paginate(per_page=num_results)
            elif entity_type == "publication":
                entities = Sources().search(query).paginate(per_page=num_results)
            else:
                continue

            entity_results = []
            for page in entities:
                for entity in page:
                    # if entity is none, continue, but why would that be? throw a warning
                    if entity is None:
                        logging.warn(f"Entity is None for {entity_type}")
                        logging.warn(f"Entity data: {entity}")
                        continue
                    try:
                        processed_result = await _process_entity(entity, entity_type)
                        if processed_result:
                            if isinstance(processed_result, Work):
                                url = processed_result.url
                                if "arxiv.org" in url or "biorxiv.org" in url or "medrxiv.org" in url or "pubmed.ncbi.nlm.nih.gov" in url:
                                    # Resolve the paper details using get_paper_details
                                    paper_details = await get_paper_details(url)
                                    if paper_details:
                                        # Update the Work object with the resolved details
                                        processed_result = Work(
                                            id=processed_result.id,
                                            work_type=processed_result.work_type,
                                            source_type=paper_details.source_type,
                                            name=paper_details.name,
                                            url=paper_details.url,
                                            abstract=paper_details.abstract,
                                            full_text=paper_details.full_text,
                                            year=paper_details.year,
                                            doi=paper_details.doi,
                                            authors=paper_details.authors,
                                            institutions=paper_details.institutions,
                                            publications=paper_details.publications,
                                        )
                            entity_results.append(processed_result)
                    except Exception as e:
                        logging.error(f"Error processing {entity_type} entity: {str(e)}")
                        logging.error(traceback.format_exc())
                        continue


                    if len(entity_results) >= num_results:
                        break

                if len(entity_results) >= num_results:
                    break

            logging.info(f"Found {len(entity_results)} results for {entity_type}")
            results.extend(entity_results)

        except pyalex.api.QueryError as e:
            logging.error(f"OpenAlex API query error for {entity_type}: {e}")
        except Exception as e:
            logging.exception(f"Unexpected error in search_openalex for {entity_type}")

    logging.info(f"Total results found: {len(results)}")
    return results



from sl_sources.doi import (
    download_from_doi,
    entity_from_doi,
    _get_clean_doi,
    search_and_download_doi,
    search_doi,
)
from sl_sources.models import (
    ENTITY_TYPES,
    SOURCE_TYPES,
    WORK_TYPES,
    Author,
    Institution,
    Publication,
    Publisher,
    SearchInput,
    Work,
)
import pytest

from sl_sources.serialization import serialize_entity_array


def test_clean_doi():
    """
    Test function for _get_clean_doi.

    This function tests the _get_clean_doi function with a set of test cases,
    checking if the function correctly cleans and extracts DOIs from various formats.
    """
    test_cases = [
        "10.1000/182",
        "https://doi.org/10.1000/182",
        "http://dx.doi.org/10.1000/182",
        "https://doi.org/10.6028/NIST.AI.100-1",
        "10.6028/NIST.AI.100-1",
    ]

    for case in test_cases:
        print(f"Input: {case}")
        print(f"Cleaned DOI: {_get_clean_doi(case)}")
        assert "http" not in _get_clean_doi(case)
        assert "https" not in _get_clean_doi(case)


@pytest.mark.asyncio
async def test_resolve_doi():
    """
    Test function for entity_from_doi.

    This function tests the entity_from_doi function with a set of test DOIs,
    checking if the function correctly resolves them and returns expected information.
    """
    test_dois = [
        "10.1016/b978-0-12-170150-5.50012-3",
        "10.7551/mitpress/5237.001.0001",
        "10.1080/15265161.2011.634487",
        "10.1002/biot.v3:12",
        "10.1016/j.biopsych.2005.02.031",
        "10.1080/2326263x.2016.1210989",
        "10.1088/1741-2560/13/2/021002",
    ]

    for doi in test_dois:
        result = await entity_from_doi(doi)
        print("result is")
        print(result)
        assert result is not None
        assert result.name
        assert result.source_type
        if result.source_type == "work":
            assert result.abstract
        if result.source_type == "publication":
            assert result.publisher
        if result.source_type == "work":
            assert result.url

    print("entity_from_doi tests passed!")


@pytest.mark.asyncio
async def test_download_from_doi():
    """
    Test function for download_from_doi.

    This function tests the download_from_doi function with a set of test DOIs,
    checking if the function correctly downloads full text content for each DOI.
    """
    test_dois = [
        "10.1016/b978-0-12-170150-5.50012-3",
        "10.7551/mitpress/5237.001.0001",
        "10.1088/1741-2560/13/2/021002",
        "10.3389/fneng.2014.00004",
        "10.1001/jama.1995.03520280069043",
        "10.1001/jama.1992.03480160079038",
        "10.1111/j.1532-5415.2010.03030.x",
        "10.1162/jocn.2009.21010",
    ]

    results = []
    for doi in test_dois:
        search_result = Work(
            name=doi,
            doi=doi,
            source_type=SOURCE_TYPES.DOI,
            url="https://doi.org/" + doi,
            work_type=WORK_TYPES.PAPER,
        )
        result = await download_from_doi(search_result)
        print("Test doi result is")
        print(result)
        results.append(result)
        if result is None:
            print(f"Unable to download from DOI: {doi}")
        else:
            assert result is not None
            # assert result.full_text
            # assert len(result.full_text) > 0

    serialize_entity_array(results, "doi_download_results.json")

    print("download_from_doi tests passed!")


@pytest.mark.asyncio
async def test_search_doi():
    """
    Test function for search_doi.

    This function tests the search_doi function with a sample query,
    checking if the function returns the expected number of results
    and if each result contains the required information.
    """
    query = "Artificial Intelligence"
    num_results = 3
    entity_types = [
        ENTITY_TYPES.WORK,
        ENTITY_TYPES.AUTHOR,
        ENTITY_TYPES.PUBLICATION,
        ENTITY_TYPES.INSTITUTION,
        ENTITY_TYPES.PUBLISHER,
    ]
    results = await search_doi(
        SearchInput(query=query, num_results=num_results, entity_types=entity_types)
    )

    serialize_entity_array(results, "doi_search_results.json")

    assert len(results) == num_results
    for result in results:
        assert result.doi
        assert result.name
        assert result.source_type

    print("search_doi test passed!")


@pytest.mark.asyncio
async def test_search_and_download_doi():
    """
    Test function for search_and_download_doi.

    This function tests the search_and_download_doi function with a sample query,
    checking if the function returns the expected number of results and if each
    result contains the required information including full text content.

    It also verifies that the linkages between entities are correctly established.
    """
    query = "Neuroscience"
    num_results = 2
    results = await search_and_download_doi(
        SearchInput(query=query, num_results=num_results)
    )

    serialize_entity_array(results, "doi_search_and_download_results.json")

    assert len(results) > 0, "No results returned"

    for result in results:
        assert isinstance(
            result, (Work, Publication, Institution, Author, Publisher)
        ), f"Unexpected result type: {type(result)}"
        assert result.doi is not None, "DOI is missing"
        assert result.name is not None, "Name is missing"

        if isinstance(result, Work):
            assert result.abstract is not None, "Abstract is missing for Work"
            assert isinstance(result.authors, list), "Authors should be a list for Work"
            assert isinstance(
                result.publications, list
            ), "Publications should be a list for Work"
            assert isinstance(
                result.institutions, list
            ), "Institutions should be a list for Work"

        elif isinstance(result, Publication):
            assert result.publisher is None or isinstance(
                result.publisher, Publisher
            ), "Publisher should be None or a Publisher instance for Publication"

    # Test linkages
    works = [r for r in results if isinstance(r, Work)]
    if works:
        work = works[0]
        for author in work.authors:
            assert isinstance(author, Author), "Linked author is not an Author instance"
        for publication in work.publications:
            assert isinstance(
                publication, Publication
            ), "Linked publication is not a Publication instance"
        for institution in work.institutions:
            assert isinstance(
                institution, Institution
            ), "Linked institution is not an Institution instance"

    publications = [r for r in results if isinstance(r, Publication)]
    if publications:
        publication = publications[0]
        if publication.publisher:
            assert isinstance(
                publication.publisher, Publisher
            ), "Linked publisher is not a Publisher instance"

    print("All assertions passed")



from typing import Dict, List, Type
import pytest

from sl_sources.google_books import (
    search_google_books,
    download_from_google_books,
    search_and_download_from_google_books,
)
from sl_sources.models import (
    ENTITY_TYPES,
    SOURCE_TYPES,
    WORK_TYPES,
    Author,
    Entity,
    Publication,
    Publisher,
    SearchInput,
    Work,
)
from sl_sources.serialization import serialize_entity_array

@pytest.mark.asyncio
async def test_search_google_books():
    query: str = "Artificial Intelligence"
    num_results: int = 5
    search_input: SearchInput = SearchInput(
        query=query,
        num_results=num_results,
        source_type=SOURCE_TYPES.GOOGLE_BOOKS,
        entity_types=[
            ENTITY_TYPES.WORK,
            ENTITY_TYPES.AUTHOR,
            ENTITY_TYPES.PUBLICATION,
            ENTITY_TYPES.PUBLISHER,
        ],
    )

    results = await search_google_books(search_input)

    serialize_entity_array(results, "google_books_search_results.json")

    assert len(results) > 0
    for result in results:
        assert isinstance(result, (Work, Author, Publication, Publisher))
        if isinstance(result, Work):
            assert result.work_type == WORK_TYPES.BOOK
            assert result.source_type == SOURCE_TYPES.GOOGLE_BOOKS

    print("search_google_books test passed!")

@pytest.mark.asyncio
async def test_download_from_google_books():
    # Use a known book ID for testing
    book_id = "zyTCAlFPjgYC"  # This is the ID for "The Hitchhiker's Guide to the Galaxy"
    work = Work(
        id=book_id,
        name="Test Book",
        url=f"https://www.googleapis.com/books/v1/volumes/{book_id}",
        source_type=SOURCE_TYPES.GOOGLE_BOOKS,
        work_type=WORK_TYPES.BOOK,
    )

    result = await download_from_google_books(work)

    serialize_entity_array([result], "google_books_download_results.json")

    assert result.full_text
    assert len(result.full_text) > 0
    assert result.abstract
    assert result.authors
    assert result.year
    assert result.publications

    print("download_from_google_books test passed!")

@pytest.mark.asyncio
async def test_search_and_download_from_google_books():
    """
    Run tests for the Google Books search and download functionality.

    This function performs a test search, processes the results, and
    prints out detailed information about the entities found.
    """
    query: str = "Artificial Intelligence"
    num_results: int = 5
    search_input: SearchInput = SearchInput(
        query=query,
        num_results=num_results,
        source_type=SOURCE_TYPES.GOOGLE_BOOKS,
        entity_types=[
            ENTITY_TYPES.WORK,
            ENTITY_TYPES.AUTHOR,
            ENTITY_TYPES.PUBLICATION,
            ENTITY_TYPES.PUBLISHER,
        ],
    )

    print(f"Searching for '{query}' with a limit of {num_results} results...")

    # Perform the search and download operation
    entities: List[Entity] = await search_and_download_from_google_books(search_input)

    serialize_entity_array(entities, "google_books_search_and_download_results.json")

    print(f"Received {len(entities)} entities.")

    # Count the occurrences of each entity type
    entity_counts: Dict[str, int] = {}
    for entity in entities:
        entity_type: str = type(entity).__name__
        entity_counts[entity_type] = entity_counts.get(entity_type, 0) + 1

    # Print the count of each entity type
    print("\nEntity types and counts:")
    for entity_type, count in entity_counts.items():
        print(f"{entity_type}: {count}")

    # Check if all required entity types are present in the results
    required_types: List[Type] = [Work, Author, Publication, Publisher]
    for required_type in required_types:
        if required_type.__name__ not in entity_counts:
            print(f"Warning: No {required_type.__name__} entities found")

    # Print detailed information for each entity
    for entity in entities:
        print(f"\n{type(entity).__name__}:")
        print(f"  ID: {entity.id}")
        print(f"  Name: {entity.name}")
        if isinstance(entity, Work):
            # Print additional details for Work entities
            print(f"  URL: {entity.url}")
            print(f"  Authors: {', '.join(author.name for author in entity.authors)}")
            print(f"  Year: {entity.year}")
            print(f"  Abstract length: {len(entity.abstract) if entity.abstract else 0}")
            print(f"  Full text length: {len(entity.full_text) if entity.full_text else 0}")
            print(f"  Publications: {', '.join(pub.name for pub in entity.publications)}")

    print("\nTest completed successfully!")


from typing import Dict

import pytest
from sl_sources.http import ThrottledClientSession

from sl_sources.models import (
    SOURCE_TYPES,
    WORK_TYPES,
    Author,
    Institution,
    Publication,
    Publisher,
    SearchInput,
    Work,
)
from sl_sources.semantic_scholar import (
    download_from_semantic_scholar,
    search_and_download_from_semantic_scholar,
    search_semantic_scholar,
)
from sl_sources.serialization import serialize_entity_array
from sl_sources.user_agent import get_user_agent_header


@pytest.mark.asyncio
async def test_search_semantic_scholar():
    query = "Artificial Intelligence"
    num_results = 2
    search_input = SearchInput(
        query=query,
        num_results=num_results,
        source_type=SOURCE_TYPES.SEMANTIC_SCHOLAR,
        entity_types=["work", "author", "publication"],
    )

    results = await search_semantic_scholar(search_input)

    serialize_entity_array(results, "semantic_scholar_search_results.json")

    for result in results:
        assert isinstance(result, (Work, Author, Publication, Institution, Publisher))

    print("search_semantic_scholar test passed!")


@pytest.mark.asyncio
async def test_download_semantic_scholar():
    url = "https://www.semanticscholar.org/paper/Attention-Is-All-You-Need-Vaswani-Shazeer/204e3073870fae3d05bcbc2f6a8e263d9b72e776"
    work = Work(
        name="Test work",
        url=url,
        source_type=SOURCE_TYPES.SEMANTIC_SCHOLAR,
        work_type=WORK_TYPES.ARTICLE,
    )

    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=get_user_agent_header()
    ) as session:
        result = await download_from_semantic_scholar(work, session)

    assert result.full_text
    assert len(result.full_text) > 0

    serialize_entity_array([result], "semantic_scholar_download_results.json")

    print("download_semantic_scholar test passed!")


@pytest.mark.asyncio
async def test_search_and_download_from_semantic_scholar():
    """
    Run a test search and download operation using Semantic Scholar.

    This function performs a sample search for "Artificial Intelligence",
    processes the results, and prints detailed information about the
    entities found.
    """
    print("Starting test for search_and_download function...")

    query = "Artificial Intelligence"
    num_results = 1
    search_input = SearchInput(
        query=query,
        num_results=num_results,
        source_type=SOURCE_TYPES.SEMANTIC_SCHOLAR,
        entity_types=["work", "author", "publication"],
    )

    print(f"Searching for '{query}' with a limit of {num_results} results...")
    entities = await search_and_download_from_semantic_scholar(search_input)

    print(f"Received {len(entities)} entities.")

    serialize_entity_array(
        entities, "semantic_scholar_search_and_download_results.json"
    )

    # Check entity types and count
    entity_counts: Dict[str, int] = {}
    for entity in entities:
        entity_type = type(entity).__name__
        entity_counts[entity_type] = entity_counts.get(entity_type, 0) + 1

    print("\nEntity types and counts:")
    for entity_type, count in entity_counts.items():
        print(f"{entity_type}: {count}")

    # Check for required entity types
    required_types = [Work, Author, Publication]
    for required_type in required_types:
        if required_type.__name__ not in entity_counts:
            print(f"Warning: No {required_type.__name__} entities found")

    # Print details of each entity
    for entity in entities:
        print(f"\n{type(entity).__name__}:")
        print(f"  ID: {entity.id}")
        print(f"  Name: {entity.name}")
        if isinstance(entity, Work):
            print(f"  URL: {entity.url}")
            print(f"  Authors: {', '.join(author.name for author in entity.authors)}")
            print(f"  Year: {entity.year}")
            print(f"  DOI: {entity.doi}")
            print(
                f"  Abstract length: {len(entity.abstract) if entity.abstract else 0}"
            )
            print(
                f"  Full text length: {len(entity.full_text) if entity.full_text else 0}"
            )

    print("\nTest completed successfully!")



from typing import Any, Dict

import aiohttp
import pytest
from sl_sources.models import Work

from sl_sources.papers import (
    get_arxiv_details,
    get_paper_details,
    get_pubmed_details,
    likely_pdf,
    pubmed_to_pdf_url,
)


@pytest.mark.asyncio
async def test_get_arxiv_details():
    """
    Test function for get_arxiv_details.

    This function tests the get_arxiv_details function with a sample arXiv URL.
    """
    url: str = "https://arxiv.org/abs/2303.08774"
    result: Work = await get_arxiv_details(url)
    print(f"ArXiv paper details: {result.model_dump()}")
    assert result.name
    assert result.authors
    assert result.abstract


@pytest.mark.asyncio
async def test_get_pubmed_details():
    """
    Test function for get_pubmed_details.

    This function tests the get_pubmed_details function with a sample PubMed URL.
    """
    url: str = "https://pubmed.ncbi.nlm.nih.gov/35580832"
    result: Work = await get_pubmed_details(url)
    print(f"PubMed paper details: {result.model_dump()}")
    assert result.name
    assert result.authors
    assert result.abstract


@pytest.mark.asyncio
async def test_pubmed_to_pdf_url():
    """
    Test function for pubmed_to_pdf_url.

    This function tests the pubmed_to_pdf_url function with a sample PubMed URL.
    """
    url: str = "https://pubmed.ncbi.nlm.nih.gov/35580832"
    async with aiohttp.ClientSession() as session:
        pdf_url: str = await pubmed_to_pdf_url(url, session)
    print(f"PubMed full-text URL: {pdf_url}")
    assert pdf_url.startswith("http")  # Check that we got a URL
    async with aiohttp.ClientSession() as session:
        async with session.get(pdf_url) as response:
            assert response.status == 200  # Check that the URL is accessible


@pytest.mark.asyncio
async def test_likely_pdf():
    """
    Test function for likely_pdf.

    This function tests the likely_pdf function with a known PDF URL.
    """
    pdf_url: str = "https://arxiv.org/pdf/2303.08774.pdf"
    async with aiohttp.ClientSession() as session:
        async with session.get(pdf_url) as response:
            is_likely_pdf: bool = await likely_pdf(response)
    print(f"Is likely PDF: {is_likely_pdf}")
    assert is_likely_pdf == True


@pytest.mark.asyncio
async def test_get_paper_details():
    """
    Test function for get_paper_details.

    This function tests the get_paper_details function with both arXiv and PubMed URLs.
    It also tests error handling for unsupported URLs.
    """
    arxiv_url: str = "https://arxiv.org/abs/2303.08774"
    pubmed_url: str = "https://pubmed.ncbi.nlm.nih.gov/35580832"
    biorxiv_url: str = "https://www.biorxiv.org/content/10.1101/2024.05.02.563883v1"

    arxiv_result: Work = await get_paper_details(arxiv_url)
    pubmed_result: Work = await get_paper_details(pubmed_url)

    print(f"ArXiv paper details: {arxiv_result.model_dump()}")
    print(f"PubMed paper details: {pubmed_result.model_dump()}")

    assert arxiv_result.id == "2303.08774"
    assert pubmed_result.id == "35580832"

    try:
        await get_paper_details("https://example.com")
    except ValueError as e:
        print(f"Expected error raised: {e}")


import pytest
from sl_sources.audio_transcriber import AudioTranscriber
import asyncio
import os

NAME_OF_PACKAGE: str = "sources"  # TODO: We probably don't want to hardcode this

# Other things to test in future, not worth testing on Aug 12, 2024:
# - Handles audio files with low sampling rates
# - Handles very short and very long audio files (or else raises error on lenghts its not designed for)
# - Does not provide text for silent audio file.

################################# Helper Functions #################################
KEY_WORD_IN_BEULLER_CLIP = "life moves pretty fast"
KEY_WORD_IN_STAR_WARS_CLIP = "Star Wars"
VALID_AUDIO_PATHS_OR_URLS = [  # Modelled as (audio_file_path_or_url, expected_keyword_in_transcription)
    (
        "tests/test_data/audio_transcriber_data/short-clip-buellers-life-moves-pretty-fast.wav",
        KEY_WORD_IN_BEULLER_CLIP,
    ),
    (
        "https://static.deepgram.com/examples/Bueller-Life-moves-pretty-fast.wav",
        KEY_WORD_IN_BEULLER_CLIP,
    ),
    (
        "tests/test_data/audio_transcriber_data/star_wars_podcast.mp3",
        KEY_WORD_IN_STAR_WARS_CLIP
    ),
    (
        "https://dcs.megaphone.fm/STU5398380055.mp3?key=20757e325545ea9f30567740f86f1cf7&request_event_id=975e70d0-8c85-4b6e-9c2a-915e1017a4cb&timetoken=1723944464_C318A1072E76C4935CEC61B4BC61AE6C",
        KEY_WORD_IN_STAR_WARS_CLIP
    ),
]


def create_transcription_tests_cases() -> list[tuple[str, str, bool]]:
    test_cases = []
    deepgram_enabled_statuses = [True, False]
    for deepgram_enabled in deepgram_enabled_statuses:
        for audio_file_path, expected_keyword in VALID_AUDIO_PATHS_OR_URLS:
            test_cases.append((audio_file_path, expected_keyword, deepgram_enabled))
    return test_cases


##################################### Tests #########################################


@pytest.mark.parametrize(
    "audio_file_path_or_url, expected_keyword, deepgram_enabled",
    create_transcription_tests_cases(),
)
def test_properly_chunks_valid_file_types(
    audio_file_path_or_url: str, expected_keyword: str, deepgram_enabled: bool
) -> None:

    if expected_keyword == KEY_WORD_IN_STAR_WARS_CLIP:
        pytest.skip(
            "The Star wars podcast takes too long to process during unit testing."
        )

    deepgram_api_key = os.getenv("DEEPGRAM_API_KEY")
    if deepgram_enabled and not deepgram_api_key:
        pytest.skip("Deepgram API key not found.")

    if not deepgram_enabled:
        del os.environ["DEEPGRAM_API_KEY"]

    try:
        if audio_file_path_or_url.startswith("http"):
            absolute_path_or_url = audio_file_path_or_url
        else:
            current_file_path = os.path.abspath(__file__)
            package_path = os.path.dirname(current_file_path)
            while os.path.basename(package_path) != NAME_OF_PACKAGE:
                package_path = os.path.dirname(package_path)
            absolute_path_or_url = str(
                os.path.join(package_path, audio_file_path_or_url)
            )

        transcription_output = asyncio.run(
            AudioTranscriber.transcribe_audio(absolute_path_or_url)
        )

        transcription_duration = transcription_output.audio_duration_in_seconds
        assert transcription_duration > 0

        transcribed_chunks = transcription_output.transcript_chunks
        expected_num_chunks = 2 if transcription_duration > 30 else 1
        assert len(transcribed_chunks) >= expected_num_chunks
        assert all([len(chunk) > 0 for chunk in transcribed_chunks])
        assert expected_keyword.lower() in "".join(transcribed_chunks).lower()

        transcription = transcription_output.transcript
        assert expected_keyword.lower() in transcription.lower()
        assert len(transcription) > 10
        assert isinstance(transcription, str)

        transcription_with_all_spaces_removed = transcription.replace(" ", "").replace(
            "\n", ""
        )
        combined_chunks_with_all_spaces_removed = (
            "".join(transcribed_chunks).replace(" ", "").replace("\n", "")
        )
        assert (
            transcription_with_all_spaces_removed
            == combined_chunks_with_all_spaces_removed
        )
    finally:
        if deepgram_api_key is not None:
            os.environ["DEEPGRAM_API_KEY"] = deepgram_api_key



from .sources import search_source, download_search_result, search_and_download
from .crawl import crawl
from sl_sources.models import (
    SOURCE_TYPES,
    ENTITY_TYPES,
    WORK_TYPES,
    INSTITUTION_TYPES,
    PUBLICATION_TYPES,
    Entity,
    Publisher,
    Author,
    Institution,
    Publication,
    Work,
    SearchInput,
    CrawlInput,
)

__all__ = [
    # Types
    "SOURCE_TYPES",
    "ENTITY_TYPES",
    "WORK_TYPES",
    "INSTITUTION_TYPES",
    "PUBLICATION_TYPES",
    "Entity",
    "Publisher",
    "Author",
    "Institution",
    "Publication",
    "Work",
    "SearchInput",
    "CrawlInput",
    # Functions
    "crawl",
    "search_source",
    "download_search_result",
    "search_and_download",
]



import asyncio
import io
import logging
import re
import traceback
from typing import List, Optional, Tuple

import aiohttp
from playwright.async_api import TimeoutError as PlaywrightTimeoutError
from playwright.async_api import async_playwright
from pypdf import PdfReader

from .google import search_google
from .http import ThrottledClientSession
from .models import (ENTITY_TYPES, INSTITUTION_TYPES, PUBLICATION_TYPES,
                     SOURCE_TYPES, WORK_TYPES, Author, Entity, Institution,
                     Publication, Publisher, SearchInput, Work,
                     create_consistent_uuid)
from .scrape import browser_scrape, get_browser_and_page
from .user_agent import get_user_agent_header

# Set up logging configuration
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


async def search_and_download_from_google_scholar(
    search_input: SearchInput,
) -> List[Entity]:
    """
    Search Google Scholar and download full text content for the search results.

    This function performs a search on Google Scholar, processes the results,
    and attempts to download the full text for each work found.

    Parameters
    ----------
    search_input : SearchInput
        An object containing search parameters such as query, number of results,
        and entity types to search for.

    Returns
    -------
    List[Entity]
        A list of Entity objects (Works, Authors, Publications) found and processed.

    Notes
    -----
    This function uses a rate-limited client session to respect Google Scholar's
    usage policies. It also handles deduplication of publications.
    """
    logging.info(f"Starting search with input: {search_input}")

    # Perform the initial search on Google Scholar
    search_results = await search_google_scholar(search_input)

    entities: List[Entity] = []
    publications_set = set()  # To keep track of unique publications

    # Create a rate-limited client session for downloading content
    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=get_user_agent_header()
    ) as session:
        for result in search_results:
            if isinstance(result, Work):
                # Process Work entities
                entities.append(result)
                # Add authors to the entities list
                for author in result.authors:
                    if author not in entities:
                        entities.append(author)
                # Add publications to the entities list, avoiding duplicates
                for publication in result.publications:
                    if publication.id not in publications_set:
                        entities.append(publication)
                        publications_set.add(publication.id)

                try:
                    # Attempt to download the full text for the work
                    result = await download_from_google_scholar(result, session)
                    entities.append(result)
                except Exception as e:
                    logging.error(f"Error downloading full text for {result.name}: {e}")
                    entities.append(result)
            elif isinstance(result, Author):
                # Process Author entities
                entities.append(result)
            elif isinstance(result, Publication):
                # Process Publication entities, avoiding duplicates
                if result.id not in publications_set:
                    entities.append(result)
                    publications_set.add(result.id)

    logging.info(f"Total entities found: {len(entities)}")
    return entities


async def search_google_scholar(search_input: SearchInput) -> List[Entity]:
    query: str = search_input.query
    num_results: int = search_input.num_results
    entity_types: List[str] = search_input.entity_types

    results: List[Entity] = []
    seen_authors = set()
    seen_institutions = set()

    async with async_playwright() as p:
        try:
            browser, page = await get_browser_and_page(p)
            search_url: str = (
                f"https://scholar.google.com/scholar?q={query}&num={num_results}"
            )
            logging.info(f"Navigating to: {search_url}")

            await page.goto(search_url, wait_until="networkidle", timeout=60000)
            await asyncio.sleep(3)

            if await page.query_selector("#captcha-form"):
                logging.error("CAPTCHA detected. Google is blocking our requests.")
                return results

            items = await page.query_selector_all("#gs_res_ccl_mid > div")
            logging.info(f"Found {len(items)} items on the page")

            for item in items:
                try:
                    title_element = await item.query_selector("h3 a")
                    name: str = (
                        await title_element.inner_text() if title_element else ""
                    )
                    url: str = (
                        await title_element.get_attribute("href")
                        if title_element
                        else ""
                    )

                    authors_year_element = await item.query_selector(".gs_a")
                    authors_year_text: str = (
                        await authors_year_element.inner_text()
                        if authors_year_element
                        else ""
                    )
                    authors, publication, year, institutions = (
                        _parse_authors_year_institutions(authors_year_text)
                    )

                    snippet_element = await item.query_selector(".gs_rs")
                    snippet: str = (
                        await snippet_element.inner_text() if snippet_element else ""
                    )

                    doi = await _extract_doi(page, url)

                    logging.info(
                        f"Extracted data: Title: {name}, Authors: {authors}, Year: {year}, "
                        f"Publication: {publication}, Institutions: {institutions}, DOI: {doi}"
                    )

                    # Create a Work entity
                    work = Work(
                        work_type=WORK_TYPES.PAPER,
                        id=create_consistent_uuid(name),
                        source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                        name=name,
                        url=url,
                        abstract=snippet,
                        full_text="",
                        year=year,
                        doi=doi,
                        authors=[],
                        institutions=[],
                    )

                    for author, author_institutions in zip(authors, institutions):
                        author_id = create_consistent_uuid(author)
                        author_entity = Author(
                            id=author_id,
                            name=author,
                            source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                            institutions=[],
                        )

                        for inst in author_institutions:
                            inst_id = create_consistent_uuid(inst)
                            institution_entity = Institution(
                                id=inst_id,
                                name=inst,
                                source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                                institution_type=INSTITUTION_TYPES.ACADEMIC,
                            )
                            author_entity.institutions.append(institution_entity)
                            if inst_id not in seen_institutions:
                                seen_institutions.add(inst_id)
                                if "institution" in entity_types:
                                    results.append(institution_entity)

                        work.authors.append(author_entity)
                        if author_id not in seen_authors:
                            seen_authors.add(author_id)
                            if "author" in entity_types:
                                results.append(author_entity)

                    if publication:
                        publisher = _extract_publisher(publication)
                        pub = Publication(
                            id=create_consistent_uuid(publication),
                            name=publication,
                            source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                            publication_type=PUBLICATION_TYPES.JOURNAL,
                            publisher=publisher,
                        )
                        work.publications.append(pub)
                        if "publication" in entity_types:
                            results.append(pub)

                    # Update the institutions attribute of the Work object
                    work.institutions = list(seen_institutions)
                    results.append(work)

                except Exception as e:
                    logging.error(f"Error processing item: {e}")
                    logging.error(f"Item HTML: {await item.inner_html()}")

        except PlaywrightTimeoutError:
            logging.error("Timeout while loading Google Scholar page")
        except Exception as e:
            logging.error(f"Unexpected error during Google Scholar search: {e}")
            logging.error(f"Traceback: {traceback.format_exc()}")
        finally:
            await browser.close()

    logging.info(f"Total results found: {len(results)}")
    return results[:num_results]


def _parse_authors_year_institutions(
    authors_year_text: str,
) -> Tuple[List[str], str, Optional[int], List[List[str]]]:
    import re

    year_match = re.search(r"\b(19|20)\d{2}\b", authors_year_text)
    if year_match:
        year = int(year_match.group())
        authors_publication = authors_year_text.replace(str(year), "").strip()
    else:
        year = None
        authors_publication = authors_year_text.strip()

    parts = authors_publication.split(" - ", 1)
    authors_part = parts[0].strip()
    publication = parts[1].strip() if len(parts) > 1 else ""

    institution_keywords = (
        r"University|Institute|College|School|Center|Laboratory|Academy|Foundation"
    )
    author_list = re.split(r",\s*(?=[^,]+(?:,\s*[^,]+)*$)", authors_part)

    authors = []
    institutions = []

    for author in author_list:
        author = author.strip()
        author_parts = author.split(", ")
        author_name = author_parts[0]
        authors.append(author_name)

        if len(author_parts) > 1:
            institution_match = re.search(
                rf"({institution_keywords}.*?)(?=$)", author_parts[1]
            )
            if institution_match:
                institutions.append([institution_match.group(1).strip()])
            else:
                institutions.append([])
        else:
            institutions.append([])

    return authors, publication, year, institutions


def _extract_publisher(publication: str) -> Optional[Publisher]:
    known_publishers = [
        "Elsevier",
        "Springer",
        "Wiley",
        "IEEE",
        "Taylor & Francis",
        "SAGE",
        "Oxford University Press",
        "Cambridge University Press",
        "Nature Publishing Group",
        "ACM",
        "ACS",
        "AIP",
        "APS",
        "ASME",
        "IOP",
        "MDPI",
        "PLOS",
        "RSC",
        "SPIE",
    ]

    try:
        for publisher_name in known_publishers:
            if publisher_name.lower() in publication.lower():
                return Publisher(
                    id=create_consistent_uuid(publisher_name),
                    name=publisher_name,
                    source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                )
    except Exception as e:
        logging.error(f"Error extracting publisher: {e}")

    return None


async def _extract_doi(page, url: str) -> Optional[str]:
    try:
        await page.goto(url, wait_until="domcontentloaded", timeout=10000)
        doi_element = await page.query_selector('meta[name="citation_doi"]')
        if doi_element:
            return await doi_element.get_attribute("content")

        content = await page.content()
        doi_match = re.search(r"\b(10\.\d{4,}(?:\.\d+)*\/\S+)\b", content)
        if doi_match:
            return doi_match.group(1)
    except Exception as e:
        logging.error(f"Error extracting DOI: {e}")

    return None


async def download_from_google_scholar(
    work: Work, session: ThrottledClientSession
) -> Work:
    """
    Download the full text content for a given work from Google Scholar.

    This function attempts to download the full text of a work using various methods,
    including direct download, Google search for PDFs, and web scraping.

    Parameters
    ----------
    work : Work
        The Work object containing information about the publication to download.
    session : ThrottledClientSession
        A rate-limited aiohttp client session for making HTTP requests.

    Returns
    -------
    Work
        The Work object with the full text content if successfully downloaded, or the original Work object if not.

    Notes
    -----
    This function tries multiple approaches to obtain the full text:
    1. Direct download from the work's URL
    2. Google search for a PDF of the work
    3. Web scraping of the work's URL
    """
    if work.type != ENTITY_TYPES.WORK:
        raise ValueError("Work is required to download full text")
    if not work.url:
        raise ValueError("Work URL is required to download full text")

    try:
        # Attempt to download and extract text from the primary URL
        full_text = await _download_and_extract_text(work.url, session)
        work.full_text = full_text
        return work
    except aiohttp.ClientResponseError as e:
        logging.error(
            f"Could not get the full text from the primary URL: {work.url}\nContinuing with other approaches..."
        )

    # If primary URL fails, search Google for a PDF
    search_results = await search_google(
        SearchInput(
            query=f'"{work.name}"',
            file_type="pdf",
            num_results=3,
            source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
            entity_types=[ENTITY_TYPES.WORK],
        )
    )
    for search_result in search_results:
        try:
            # Attempt to download and extract text from each search result
            text = await _download_and_extract_text(search_result.url, session)
            if text:
                work.url = search_result.url
                work.full_text = text
                return work
        except Exception as e:
            logging.error(f"Error downloading PDF from Google search: {e}")

    try:
        # If all else fails, attempt to scrape the text from the webpage
        text = await browser_scrape(work.url)
        if text:
            work.url = work.url
            work.full_text = text
            return work
    except Exception as e:
        logging.error(f"Error scraping full text: {e}")

    # If no method succeeds, return an empty string
    raise ValueError("Could not download full text")


async def _download_and_extract_text(url: str, session: aiohttp.ClientSession) -> str:
    """
    Download content from a URL and extract text, handling both PDFs and web pages.

    This function attempts to download content from a given URL and extract
    text from it. It can handle both PDF files and regular web pages.

    Parameters
    ----------
    url : str
        The URL to download content from.
    session : aiohttp.ClientSession
        An aiohttp client session for making HTTP requests.

    Returns
    -------
    str
        The extracted text content from the URL.

    Notes
    -----
    For PDF files, this function uses pypdf to extract text.
    For web pages, it uses the browser_scrape function for extraction.
    """
    async with session.get(url, allow_redirects=True) as response:
        if response.status == 200:
            content_type = response.headers.get("Content-Type", "").lower()
            if "application/pdf" in content_type:
                # Handle PDF content
                pdf_content = await response.read()
                pdf_file = io.BytesIO(pdf_content)
                try:
                    # Use pypdf to read the PDF and extract text
                    pdf_reader = PdfReader(pdf_file)
                    text = "\n".join(page.extract_text() for page in pdf_reader.pages)
                    return text
                except Exception as e:
                    logging.error(f"Error extracting text from PDF: {e}")
            else:
                # For non-PDF content, use browser_scrape to extract text
                return await browser_scrape(url)
        else:
            # Log a warning if the download fails
            logging.warning(
                f"Failed to download content from {url}. Status code: {response.status}"
            )
            return None



import logging
import os
import tempfile
import aiohttp
import librosa
import numpy as np
from pydantic import BaseModel
from tqdm import tqdm
from sl_ai_models import DeepgramNova2
from sl_ai_models.deepgram_nova2 import DeepgramOutput
import asyncio

logger = logging.getLogger(__name__)

import librosa
from tqdm import tqdm


def transcribe(audio_file_path_or_audio_url: str) -> str:
    transcription_output = asyncio.run(
        AudioTranscriber.transcribe_audio(audio_file_path_or_audio_url)
    )
    transcription = transcription_output.transcript
    return transcription


class AudioTranscriptionOutput(BaseModel):
    transcript_chunks: list[str]
    audio_duration_in_seconds: float
    other_metadata: dict | None = None

    @property
    def transcript(self) -> str:
        return "\n\n".join(self.transcript_chunks)


class AudioTranscriber:
    _TARGET_RESAMPLING_RATE_IN_HZ = 16000

    @classmethod
    async def transcribe_audio(
        cls, audio_file_absolute_path_or_url: str
    ) -> AudioTranscriptionOutput:
        logger.info(f"Transcribing audio file '{audio_file_absolute_path_or_url}'.")
        input_is_url = audio_file_absolute_path_or_url.startswith("http")
        input_is_file = os.path.exists(audio_file_absolute_path_or_url)
        assert (
            input_is_url or input_is_file
        ), f"Input '{audio_file_absolute_path_or_url}' is not a valid URL or file path."

        deepgram_enabled = os.getenv("DEEPGRAM_API_KEY") is not None
        if deepgram_enabled:
            transcription_chunks, deepgram_output = (
                await cls.__audio_to_chunked_transcript_with_deepgram(
                    audio_file_absolute_path_or_url
                )
            )
            audio_duration_in_seconds = deepgram_output.audio_duration_in_seconds
            other_metadata = deepgram_output.model_dump()
        elif input_is_url:
            transcription_chunks, duration = (
                await cls.__audio_url_to_chunked_transcript_with_whisper(
                    audio_file_absolute_path_or_url
                )
            )
            audio_duration_in_seconds = duration
            other_metadata = None
        else:
            transcription_chunks, duration = (
                await cls.__audio_file_path_to_chunked_transcript_with_whisper(
                    audio_file_absolute_path_or_url
                )
            )
            audio_duration_in_seconds = duration
            other_metadata = None

        output = AudioTranscriptionOutput(
            transcript_chunks=transcription_chunks,
            audio_duration_in_seconds=audio_duration_in_seconds,
            other_metadata=other_metadata,
        )
        logger.debug(
            f"Transcription complete for audio '{audio_file_absolute_path_or_url}'."
        )
        logger.debug(f"Transcription output: {output}")
        return output

    @classmethod
    async def __audio_to_chunked_transcript_with_deepgram(
        cls, audio_file_aboslute_path: str
    ) -> tuple[list[str], DeepgramOutput]:
        logger.debug(
            f"Transcribing audio file '{audio_file_aboslute_path}' with Deepgram."
        )
        model = DeepgramNova2()
        transcription_output = await model.invoke(audio_file_aboslute_path)
        transcript = transcription_output.transcript
        transcription_chunks = [chunk for chunk in transcript.split("\n") if chunk]
        return transcription_chunks, transcription_output

    @classmethod
    async def __audio_url_to_chunked_transcript_with_whisper(
        cls, audio_url: str
    ) -> tuple[list[str], float]:
        assert audio_url.startswith("http"), f"Input '{audio_url}' is not a valid URL."
        logger.debug(f"Trying to get audio file from '{audio_url}'.")
        async with aiohttp.ClientSession() as session:
            async with session.get(audio_url) as response:
                response.raise_for_status()
                with tempfile.NamedTemporaryFile(
                    suffix=".mp3", delete=True
                ) as temp_file:
                    temp_file.write(await response.read())
                    temp_file.flush()
                    logger.debug(
                        f"Got audio file for '{audio_url}'. Starting transcription"
                    )
                    transcription_chunks, duration = (
                        await cls.__audio_file_path_to_chunked_transcript_with_whisper(
                            temp_file.name
                        )
                    )
                    return transcription_chunks, duration

    @classmethod
    async def __audio_file_path_to_chunked_transcript_with_whisper(
        cls, audio_file_aboslute_path: str
    ) -> tuple[list[str], float]:
        assert os.path.exists(audio_file_aboslute_path), f"Input '{audio_file_aboslute_path}' is not a valid file path."
        logger.debug(f"Transcribing audio file '{audio_file_aboslute_path}' with Whisper.")
        audio_data = (
            cls.__load_audio_file_as_amplitude_array_using_target_sampling_rate(
                audio_file_aboslute_path
            )
        )
        chunk_length_in_seconds = 30
        audio_data_chunks = cls.__chunk_audio_data_into_equal_pieces(
            audio_data, chunk_length_in_seconds
        )
        duration = len(audio_data_chunks) * chunk_length_in_seconds
        transcription_chunks = await cls.__transcribe_audio_chunks_with_whisper(
            audio_data_chunks
        )
        return transcription_chunks, duration

    @classmethod
    def __load_audio_file_as_amplitude_array_using_target_sampling_rate(
        cls, audio_file_path: str
    ) -> np.ndarray:
        audio_data, sampling_rate = librosa.load(audio_file_path)
        resampled_audio = librosa.resample(
            audio_data,
            orig_sr=sampling_rate,
            target_sr=AudioTranscriber._TARGET_RESAMPLING_RATE_IN_HZ,
        )
        return resampled_audio

    @classmethod
    def __chunk_audio_data_into_equal_pieces(
        cls, audio_data: np.ndarray, chunk_length_in_seconds: int
    ) -> list[np.ndarray]:
        chunk_length_samples = (
            chunk_length_in_seconds * cls._TARGET_RESAMPLING_RATE_IN_HZ
        )
        audio_chunks = [
            audio_data[i : i + chunk_length_samples]
            for i in range(0, len(audio_data), chunk_length_samples)
        ]
        return audio_chunks

    @classmethod
    async def __transcribe_audio_chunks_with_whisper(
        cls, audio_chunks: list[np.ndarray]
    ) -> list[str]:
        from transformers import WhisperForConditionalGeneration, WhisperProcessor
        from transformers.generation.utils import GenerateOutput

        processor = WhisperProcessor.from_pretrained("openai/whisper-small.en")
        assert isinstance(processor, WhisperProcessor)

        model = WhisperForConditionalGeneration.from_pretrained(
            "openai/whisper-small.en"
        )

        transcription_chunks = []
        for chunk in tqdm(audio_chunks, desc="Transcribing"):
            inputs = processor(
                chunk,
                sampling_rate=cls._TARGET_RESAMPLING_RATE_IN_HZ,
                return_tensors="pt",
            )

            if "input_features" in inputs:
                input_features = inputs["input_features"]
            else:
                raise ValueError(
                    "The processor output does not contain 'input_features'."
                )

            output = model.generate(
                input_features,
                output_scores=False,
                return_dict_in_generate=True,
                output_attentions=False,
            )
            assert isinstance(output, GenerateOutput)

            for sequence in output.sequences:
                chunk_text = processor.decode(sequence, skip_special_tokens=True)
                transcription_chunks.append(chunk_text)

        return transcription_chunks



from typing import List
from sl_sources.models import (
    ENTITY_TYPES,
    SOURCE_TYPES,
    Author,
    Institution,
    Publication,
    SearchInput,
    Work,
)
from sl_sources.sources import (
    download_search_result,
    search_and_download,
    search_source,
)
import pytest


@pytest.mark.asyncio
async def test_search_sources():
    query = "Artificial Intelligence"
    num_results = 2
    search_input = SearchInput(
        query=query,
        num_results=num_results,
        source_type=SOURCE_TYPES.OPENALEX,
        entity_types=[ENTITY_TYPES.WORK],
    )

    results = await search_source(search_input)

    # assert len(results) == num_results
    # for result in results:
    #     assert isinstance(result, Work)

    print("search_sources test passed!")


@pytest.mark.asyncio
async def test_download_search_result():
    url = "https://arxiv.org/abs/1706.03762"
    work = Work(name=url, url=url, source_type=SOURCE_TYPES.GOOGLE_SCHOLAR)

    result = await download_search_result(work)

    assert result.full_text
    assert len(result.full_text) > 0

    print("download_search_result test passed!")


@pytest.mark.asyncio
async def test_sources():
    """
    Run a series of tests to ensure the search_and_download function is working correctly
    for various sources and search inputs.
    """
    # Define test cases with different queries, result counts, and entity types
    test_cases = [
        {
            "query": "Artificial Intelligence",
            "num_results": 5,
            "entity_types": ["work", "author", "institution", "publication"],
        },
        {
            "query": "Neuroscience AND Consciousness",
            "num_results": 3,
            "entity_types": ["work", "author"],
        },
        # Additional test cases can be added here
    ]

    sources: List[SOURCE_TYPES] = [
        SOURCE_TYPES.OPENALEX,
        SOURCE_TYPES.GOOGLE_SCHOLAR,
        SOURCE_TYPES.DOI,
        SOURCE_TYPES.SEMANTIC_SCHOLAR,
        SOURCE_TYPES.VIDEO,
        SOURCE_TYPES.GOOGLE,
        # SOURCE_TYPES.TWITTER,
    ]

    # Test each source separately
    for source in sources:
        print(f"\nTesting {source}...")

        for test_case in test_cases:
            print(f"\nSearch query: {test_case['query']}")

            # Create a SearchInput object for each test case
            search_input = SearchInput(
                query=test_case["query"],
                num_results=test_case["num_results"],
                source_type=source,
                entity_types=test_case["entity_types"],
            )

            # Perform the search and download
            entities = await search_and_download(search_input)

            # Analyze and validate the results
            entity_counts = {}
            for entity in entities:
                entity_type = type(entity).__name__
                entity_counts[entity_type] = entity_counts.get(entity_type, 0) + 1

                # Perform type-specific checks
                if isinstance(entity, Work):
                    print("entity is", entity)
                    assert entity.name, "Work entity should have a name"
                    assert entity.url, "Work entity should have a URL"
                    assert (
                        entity.authors is not None
                    ), "Work entity should have authors (even if empty list)"
                elif isinstance(entity, Author):
                    assert entity.name, "Author entity should have a name"
                elif isinstance(entity, Institution):
                    assert entity.name, "Institution entity should have a name"
                elif isinstance(entity, Publication):
                    assert entity.name, "Publication entity should have a name"

            print("Entity counts:")
            for entity_type, count in entity_counts.items():
                print(f"{entity_type}: {count}")

            # Uncomment the following to check if all requested entity types are present
            # for entity_type in test_case["entity_types"]:
            #     assert entity_type.capitalize() in entity_counts, f"Requested entity type {entity_type} not found in results"

        print(f"{source} tests passed!")



import asyncio
import hashlib
import json
import os
import re
import time
import traceback
from typing import Any, Dict, List, Optional, Union

import aiofiles
import aiohttp

from sl_sources.video import can_download_with_ytdlp

from .models import (
    ENTITY_MODELS,
    ENTITY_TYPES,
    SOURCE_TYPES,
    WORK_TYPES,
    Author,
    Entity,
    Institution,
    Publication,
    Publisher,
    SearchInput,
    Work,
)

from .http import cloud_function_request, resolve_url, validate_url
from .papers import get_paper_details
from .sources import download_search_result, search_and_download, search_source


class CacheManager:
    def __init__(self, cache_filename: str):
        self.cache_filename = cache_filename
        self.entities = {}
        self.searches = {}  # Modified to be a dictionary of dictionaries, keyed by source type
        self.load_cache()
        self.lock = asyncio.Lock()

    def load_cache(self) -> None:
        if os.path.exists(self.cache_filename):
            with open(self.cache_filename, "r") as f:
                cache_data = json.load(f)
                self.entities = cache_data.get("entities", {})
                self.searches = cache_data.get("searches", {})

    async def update_cache(
        self,
        entity: Optional[Entity] = None,
        search_keyword: Optional[str] = None,
        search_results: Optional[List[Work]] = None,
        search_source_type: Optional[str] = None,  # Added search source type parameter
    ) -> None:
        async with self.lock:
            if entity is not None:
                url_hash = self.get_entity_hash(entity)
                self.entities[url_hash] = (
                    entity.model_dump() if hasattr(entity, "model_dump") else entity
                )

            if search_keyword is not None and search_results is not None and search_source_type is not None:
                if search_source_type not in self.searches:
                    self.searches[search_source_type] = {}
                self.searches[search_source_type][search_keyword] = [
                    result.model_dump() if hasattr(result, "model_dump") else result
                    for result in search_results
                ]

            await self.save_cache()

    async def save_cache(self) -> None:
        # if self.cache_filename exists, copy it to self.cache_filename.backup.json
        if os.path.exists(self.cache_filename):
            os.copy(self.cache_filename, f"{self.cache_filename.replace('.json', '')}.backup.json")
        async with aiofiles.open(self.cache_filename, "w") as f:
            cache_data = {"entities": self.entities, "searches": self.searches}
            await f.write(json.dumps(cache_data, indent=2))
            print(f"**** Cache saved to {self.cache_filename}")

    def get_entity_hash(self, entity: Entity) -> str:
        return (
            hashlib.md5(entity.url.encode()).hexdigest()
            if hasattr(entity, "url")
            else entity.id
        )


async def crawl(
    keywords: List[str] = [],
    urls: List[str] = [],
    source_types: List[str] = [SOURCE_TYPES.OPENALEX, SOURCE_TYPES.GOOGLE_SCHOLAR],
    research_topic: str = "",
    max_concurrent_threads: int = 3,
    use_cloud_function: bool = False,
    entity_types: List[str] = [
        ENTITY_TYPES.WORK,
        ENTITY_TYPES.AUTHOR,
        ENTITY_TYPES.INSTITUTION,
        ENTITY_TYPES.PUBLICATION,
    ],
    num_results: int = 5,
    cache_filename: str = "manifest.json",
    max_depth: int = 3,
    rate_limit: float = 1.0,
) -> List[Entity]:
    print(f"Crawling with {len(keywords)} keywords and {len(urls)} URLs")

    # Validate and deduplicate URLs
    validated_urls: List[str] = []
    for url in urls:
        try:
            validated_url = validate_url(url)
            validated_urls.append(validated_url)
        except ValueError as e:
            print(f"Skipping invalid URL: {url}. Error: {str(e)}")
            continue

    urls: set = set(validated_urls)
    keywords: set = set(keywords)

    # Normalize and categorize URLs
    arxiv_urls: List[str] = [url for url in urls if "arxiv.org/" in url]
    biorxiv_urls: List[str] = [url for url in urls if "biorxiv.org/" in url]
    pubmed_urls: List[str] = [url for url in urls if "pubmed.ncbi.nlm.nih.gov/" in url]
    media_urls: List[str] = [url for url in urls if can_download_with_ytdlp(url)]
    doi_urls: List[str] = [url for url in urls if "doi.org/" in url]

    # Remove categorized URLs from the main list
    urls = [
        url
        for url in urls
        if "arxiv.org" not in url
        and "biorxiv.org" not in url
        and "pubmed.ncbi.nlm.nih.gov/" not in url
        and "doi.org/" not in url
        and not can_download_with_ytdlp(url)
    ]

    url_sources: List[Work] = []

    # Create Work objects for each type of URL
    for url in doi_urls:
        id: str = url.split("doi.org/")[1]
        url_sources.append(
            Work(
                id=id,
                name=f"DOI: {id}",
                url=f"https://doi.org/{id}",
                source_type=SOURCE_TYPES.DOI,
                work_type=WORK_TYPES.PAPER,
            )
        )
    for url in arxiv_urls:
        id: str = url.split("/")[-1]
        url_sources.append(
            Work(
                id=id,
                name=f"arXiv: {id}",
                url=url,
                source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                work_type=WORK_TYPES.PAPER,
            )
        )
    for url in biorxiv_urls:
        id: str = re.search(r"(?:biorxiv\.org/content/)(.+)", url).group(1)
        url_sources.append(
            Work(
                id=id,
                name=f"bioRxiv: {id}",
                url=url,
                source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                work_type=WORK_TYPES.PAPER,
            )
        )
    for url in pubmed_urls:
        id: str = url.split("/")[-1]
        url_sources.append(
            Work(
                id=id,
                name=f"PubMed: {id}",
                url=url,
                source_type=SOURCE_TYPES.GOOGLE_SCHOLAR,
                work_type=WORK_TYPES.PAPER,
            )
        )
    for url in media_urls:
        source_type = SOURCE_TYPES.VIDEO
        id = url
        if "youtube" in id:
            id = id.split("v=")[-1].split("&")[0]
        elif "vimeo" in id:
            id = id.split("/")[-1].split("?")[0]
        elif "dailymotion" in id:
            id = id.split("/")[-1].split("?")[0]

        url_sources.append(
            Work(
                id=id,
                name=f"Media: {id}",
                url=url,
                source_type=source_type,
                work_type=WORK_TYPES.VIDEO,
            )
        )

    # Create a cache manager instance
    cache_manager = CacheManager(cache_filename)

    # Create queues for search and download tasks
    search_queue = asyncio.Queue()
    download_queue = asyncio.Queue()

    # Create semaphores for search and download operations
    search_semaphore = asyncio.Semaphore(max_concurrent_threads)
    download_semaphore = asyncio.Semaphore(max_concurrent_threads)

    # Create a list to store the results
    results: List[Entity] = []

    # Create search worker tasks
    search_workers = [
        asyncio.create_task(
            search_worker(
                search_queue,
                download_queue,
                search_semaphore,
                use_cloud_function,
                rate_limit,
                cache_manager,
            )
        )
        for _ in range(max_concurrent_threads)
    ]

    # Create download worker tasks
    download_workers = [
        asyncio.create_task(
            download_worker(
                download_queue,
                download_semaphore,
                use_cloud_function,
                research_topic,
                entity_types,
                cache_manager,
                results,
                max_depth,
                rate_limit,
            )
        )
        for _ in range(max_concurrent_threads)
    ]

    # Add initial search tasks to the queue
    for source_type in source_types:
        for keyword in keywords:
            if keyword in cache_manager.searches:
                # Use cached search results if available
                cached_results = cache_manager.searches[keyword]
                for result in cached_results:
                    await download_queue.put(result)
            else:
                # Add search task to the queue
                await search_queue.put(
                    SearchInput(
                        entity_types=entity_types,
                        query=keyword,
                        num_results=num_results,
                        source_type=source_type,
                    )
                )

    # Add initial URL sources to the download queue
    for source in url_sources:
        await download_queue.put(source)

    # Wait for all search tasks to complete
    await search_queue.join()

    # Wait for all download tasks to complete
    await download_queue.join()

    # Cancel worker tasks
    for worker in search_workers + download_workers:
        worker.cancel()

    # Wait for all worker tasks to be cancelled
    await asyncio.gather(*search_workers, *download_workers, return_exceptions=True)

    print(
        f"Crawl completed. Check the 'downloaded_data' directory {cache_filename} and for results."
    )

    # Return the results
    return results


async def search_worker(
    search_queue,
    download_queue,
    semaphore,
    use_cloud_function,
    rate_limit,
    cache_manager,
):
    while True:
        search_input = await search_queue.get()
        try:
            async with semaphore:
                search_source_type = search_input.source_type
                if search_source_type in cache_manager.searches and search_input.query in cache_manager.searches[search_source_type]:
                    # Use cached search results if available for the specific source type
                    cached_results = cache_manager.searches[search_source_type][search_input.query]
                    for result in cached_results:
                        await download_queue.put(result)
                else:
                    # Perform a new search
                    results = await _search_and_add_to_queue(
                        search_input, use_cloud_function
                    )
                    for result in results:
                        await download_queue.put(result)
                    # Update the cache with the new search results for the specific source type
                    await cache_manager.update_cache(
                        search_keyword=search_input.query, search_results=results, search_source_type=search_source_type
                    )
        except Exception as e:
            print(f"Error in search worker: {e}")
        finally:
            search_queue.task_done()
        await asyncio.sleep(rate_limit)


async def download_worker(
    download_queue,
    semaphore,
    use_cloud_function,
    research_topic,
    entity_types,
    cache_manager,
    results,
    max_depth,
    rate_limit,
):
    while True:
        item = await download_queue.get()
        try:
            if isinstance(item, dict):
                item = Work(**item)

            result = None
            async with semaphore:
                result = await _download_and_evaluate(
                    item,
                    0,
                    max_depth,
                    use_cloud_function,
                    research_topic,
                    entity_types,
                    cache_manager,
                )

            if result is not None:
                results.append(result)
                await cache_manager.update_cache(entity=result)

                if result.relevant:
                    await _crawl_links(
                        result.links,
                        depth=1,
                        max_depth=max_depth,
                        cache_manager=cache_manager,
                        parent_source=item,
                        use_cloud_function=use_cloud_function,
                        research_topic=research_topic,
                        entity_types=entity_types,
                    )
        except Exception as e:
            print(f"Error in download worker: {e}")
            traceback.print_exc()
        finally:
            download_queue.task_done()
        await asyncio.sleep(rate_limit)


async def _search_and_add_to_queue(
    search_input: SearchInput, use_cloud_function: bool
) -> List[Work]:
    if use_cloud_function:
        results = await cloud_function_request(
            "search_and_download",
            {
                "source_types": [search_input.source_type],
                "query": search_input.query,
                "num_results": search_input.num_results,
                "entity_types": search_input.entity_types,
            },
        )
    else:
        results = await search_source(search_input)

    return results


async def _download_and_evaluate(
    source: Work,
    depth: int,
    max_depth: int,
    use_cloud_function: bool,
    research_topic: str,
    entity_types: List[str],
    cache_manager: CacheManager,
) -> Entity:
    url: str = source.url
    print(f"Starting download_and_evaluate for {url}")

    try:
        url = validate_url(url)
        if "semanticscholar.org" in url:
            source.source_type = SOURCE_TYPES.SEMANTIC_SCHOLAR
    except ValueError as e:
        print(f"Skipping invalid URL: {url}. Error: {str(e)}")
        return source

    url_hash: str = cache_manager.get_entity_hash(source)

    if url_hash in cache_manager.entities:
        print(f"Using cached content for {url}")
        cached_entity = cache_manager.entities[url_hash]
        if isinstance(cached_entity, dict):
            return Work(**cached_entity)
        return cached_entity

    print(f"Downloading {url}")
    if use_cloud_function:
        entity_data = await cloud_function_request(
            "download_search_result",
            {"search_result": source.model_dump_json()},
        )
        type = entity_data.get("type")
        entity = ENTITY_MODELS[type](**entity_data)
    else:
        print("Downloading search result")
        entity = await download_search_result(source)

    if not entity:
        print(f"No content downloaded for {url}")
        entity = Work(
            id=url_hash,
            url=url,
            name=f"No content: {url}",
            abstract="No content downloaded",
            source_type=SOURCE_TYPES.CRAWL,
            work_type=WORK_TYPES.UNKNOWN,
        )
        await cache_manager.update_cache(entity=entity)
        return entity

    text = entity.full_text if entity.full_text else ""
    print(f"Downloaded content for {url}: {str(text)[:100]}...")

    print(f"Evaluating page for {url}")
    result: Dict[str, Any] = await _evaluate_page(text, research_topic)
    print(f"Evaluation result for {url}: {result}")

    if isinstance(entity, Work) and entity.type == ENTITY_TYPES.WORK:
        entity.abstract = result["abstract"]
        entity.relevant = result["relevant"]
        entity.full_text = text
        entity.links = result["links"]

    await cache_manager.update_cache(entity=entity)
    print(f"Added to cache: {url}")

    return entity


def read_cache(cache_filename: str) -> Dict[str, Any]:
    """
    Read the cache from a JSON file.
    """
    # Load or create cache
    cache: Dict[str, Any] = {}
    if os.path.exists(cache_filename):
        with open(cache_filename, "r") as f:
            cache = json.load(f)
    return cache


async def process_url(
    url: str,
    use_cloud_function: bool,
    research_topic: str,
    entity_types: List[str],
    cache: Dict[str, Any],
    cache_filename: str,
    max_depth: int,
    depth: int = 0,
) -> Optional[Entity]:
    print(f"Processing URL: {url}")

    try:
        # Validate the URL
        url = validate_url(url)
        source_type = SOURCE_TYPES.CRAWL
        if "semanticscholar.org" in url:
            source_type = SOURCE_TYPES.SEMANTIC_SCHOLAR
    except ValueError as e:
        print(f"Skipping invalid URL: {url}. Error: {str(e)}")
        return None

    # Generate a unique hash for the URL
    url_hash: str = hashlib.md5(url.encode()).hexdigest()

    # Check if the URL is already in the cache
    if url_hash in cache:
        print(f"Using cached content for {url}")
        cached_entity = cache[url_hash]
        if isinstance(cached_entity, dict):
            # Convert dict to Work entity if necessary
            return Work(**cached_entity)
        return cached_entity

    try:
        print(f"Downloading {url}")
        if use_cloud_function:
            # Use cloud function to download content
            entity_data = await cloud_function_request(
                "download_search_result",
                {
                    "search_result": Work(
                        url=url, source_type=source_type
                    ).model_dump_json()
                },
            )
            type = entity_data.get("type")
            entity = ENTITY_MODELS[type](**entity_data)
        else:
            # Download search result locally
            entity = await download_search_result(
                Work(url=url, source_type=source_type)
            )

        if not entity:
            print(f"No content downloaded for {url}")
            # Create a placeholder Work entity if no content was downloaded
            entity = Work(
                id=url_hash,
                url=url,
                name=f"No content: {url}",
                abstract="No content downloaded",
                source_type=source_type,
                work_type=WORK_TYPES.UNKNOWN,
            )
            # Update the cache with the placeholder entity
            await update_cache(cache, entity, cache_filename)
            return entity

        # Extract full text
        text = entity.full_text if entity.full_text else ""

        print(f"Downloaded content for {url}: {str(text)[:100]}...")

        print(f"Evaluating page for {url}")
        # Evaluate the page content for relevance
        result: Dict[str, Any] = await _evaluate_page(text, research_topic)
        print(f"Evaluation result for {url}: {result}")

        # Update the entity with evaluation results
        if isinstance(entity, Work) and entity.type == ENTITY_TYPES.WORK:
            entity.abstract = result["abstract"]
            entity.relevant = result["relevant"]
            entity.full_text = text
            entity.links = result["links"]

        # Update the cache with the new entity
        await update_cache(cache, entity, cache_filename)
        print(f"Added to cache: {url}")

        if result["relevant"] and depth < max_depth:
            # If the content is relevant and we haven't reached max depth, crawl the links found in the page
            for link in result["links"]:
                await process_url(
                    link,
                    use_cloud_function,
                    research_topic,
                    entity_types,
                    cache,
                    cache_filename,
                    max_depth,
                    depth + 1,
                )

        return entity

    except Exception as e:
        print(f"Error processing {url}: {e}")
        print(traceback.format_exc())
        # Create an error Work entity if an exception occurs
        error_entity = Work(
            id=url_hash,
            url=url,
            name=f"Error: {url}",
            abstract=str(e),
            source_type=source_type,
            work_type=WORK_TYPES.UNKNOWN,
        )
        # Update the cache with the error entity
        await update_cache(cache, error_entity, cache_filename)
        return error_entity


# Helper functions


def read_cache(cache_filename: str) -> Dict[str, Any]:
    """
    Read the cache from a JSON file.
    """
    import json
    import os

    cache: Dict[str, Any] = {}
    if os.path.exists(cache_filename):
        with open(cache_filename, "r") as f:
            cache = json.load(f)
    return cache


async def update_cache(
    cache: Dict[str, Any],
    entity: Optional[Entity] = None,
    search_keyword: Optional[str] = None,
    search_results: Optional[List[Work]] = None,
    filename: str = "manifest.json",
) -> None:
    async with aiofiles.open(filename, "r") as f:
        cache_data = json.loads(await f.read())

    if entity is not None:
        url_hash = (
            hashlib.md5(entity.url.encode()).hexdigest()
            if hasattr(entity, "url")
            else entity.id
        )
        cache_data["entities"][url_hash] = (
            entity.model_dump() if hasattr(entity, "model_dump") else entity
        )

    if search_keyword is not None and search_results is not None:
        cache_data["searches"][search_keyword] = [
            result.model_dump() if hasattr(result, "model_dump") else result
            for result in search_results
        ]

    async with aiofiles.open(filename, "w") as f:
        await f.write(json.dumps(cache_data, indent=2))
        print(f"Cache written to {filename}")


async def _evaluate_page(
    text: str, research_topic: str, model: str = "gpt-4o-mini"
) -> Dict[str, Any]:
    OPENAI_API_KEY: Optional[str] = os.getenv(
        "OPENAI_API_KEY", os.getenv("SOCIETY_API_KEY")
    )
    if not OPENAI_API_KEY:
        raise ValueError("Please set the OPENAI_API_KEY environment variable")

    # Truncate the text if it's too long
    max_length: int = 128000
    if len(text) > max_length:
        text = text[:max_length] + "..."

    # Construct the prompt for the OpenAI API
    prompt: str = f"""
    Analyze the following text and determine if it's relevant. 
    Also extract any URLs mentioned in the text that seem relevant to these topics.
    
    Text:
    {text}
    
    We are researching the following topic and related domains:
    {research_topic}

    Please evaluate if the text above contains relevant and substantive information for our research.

    Respond with a JSON object containing two fields:
    1. "relevant": a boolean indicating if the text is relevant
    2. "links": a list of relevant URLs extracted from the text which are worth looking at. Ignore links that are not relevant to the research topic.
    3. "abstract": a summary of the text, focusing on the most relevant information for the research topic.
    
    Example response:
    {{
        "relevant": true,
        "links": ["https://example.com/ai-article", "https://example.org/ml-study"],
        "abstract": "A summary of the text, focusing on the most relevant information for the research topic."
    }}
    """

    # Make a request to the OpenAI API
    async with aiohttp.ClientSession() as session:
        async with session.post(
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Authorization": f"Bearer {OPENAI_API_KEY}",
                "Content-Type": "application/json",
            },
            json={
                "model": model,
                "messages": [{"role": "user", "content": prompt}],
                "temperature": 0.2,
            },
        ) as response:
            if response.status == 200:
                result: Dict[str, Any] = await response.json()
                try:
                    # Parse the API response
                    evaluation: str = result["choices"][0]["message"]["content"]
                    evaluation = evaluation.replace("```json\n", "").replace(
                        "\n```", ""
                    )
                    evaluation_dict: Dict[str, Any] = json.loads(evaluation)
                    evaluation_dict["text"] = text
                    return evaluation_dict
                except json.JSONDecodeError:
                    print(
                        f"Error parsing GPT-4 response: {result['choices'][0]['message']['content']}"
                    )
                    return {
                        "relevant": False,
                        "links": [],
                        "abstract": "",
                        "text": text,
                    }
            else:
                print(f"Error calling OpenAI API: {response.status}")
                return {"relevant": False, "links": [], "abstract": "", "text": text}


async def _crawl_links(
    links: List[str],
    depth: int,
    max_depth: int,
    cache_manager: CacheManager,
    parent_source: Optional[Union[Work, Dict[str, Any]]] = None,
    use_cloud_function: bool = False,
    research_topic: Optional[str] = None,
    entity_types: Optional[List[str]] = None,
) -> List[Dict[str, Any]]:
    print(f"Crawling links at depth {depth}")
    if depth > max_depth:
        return []

    source_type: str = (
        parent_source.source_type
        if isinstance(parent_source, Work)
        else SOURCE_TYPES.CRAWL
    )

    tasks: List[asyncio.Task] = []
    for link in links:
        try:
            if isinstance(parent_source, Work):
                resolved_url: str = resolve_url(link, base_url=parent_source.url)
            elif isinstance(parent_source, dict) and "url" in parent_source:
                resolved_url: str = resolve_url(link, base_url=parent_source["url"])
            else:
                resolved_url: str = resolve_url(link)
        except ValueError as e:
            print(f"Skipping invalid URL: {link}. Error: {str(e)}")
            url_hash: str = cache_manager.get_entity_hash(
                Work(url=link, source_type=source_type, name=f"Invalid URL: {link}")
            )
            error_entity = Work(
                id=url_hash,
                url=link,
                name="Invalid URL",
                abstract=str(e),
                source_type=source_type,
                work_type=WORK_TYPES.UNKNOWN,
                authors=[],
                relevant=False,
            )
            await cache_manager.update_cache(entity=error_entity)
            continue

        url_hash: str = cache_manager.get_entity_hash(
            Work(
                url=resolved_url,
                source_type=source_type,
                name=f"Crawled URL: {resolved_url}",
            )
        )

        if url_hash in cache_manager.entities:
            print(f"URL {resolved_url} already processed or in cache")
            continue

        task: asyncio.Task = asyncio.create_task(
            _process_link(
                resolved_url,
                depth,
                max_depth,
                source_type=source_type,
                parent_source=parent_source,
                use_cloud_function=use_cloud_function,
                research_topic=research_topic,
                entity_types=entity_types,
                cache_manager=cache_manager,
            )
        )
        tasks.append(task)

    results: List[Optional[Dict[str, Any]]] = await asyncio.gather(
        *tasks, return_exceptions=True
    )
    return [
        result
        for result in results
        if result is not None and not isinstance(result, Exception)
    ]


async def _process_link(
    url: str,
    depth: int,
    max_depth: int,
    source_type: str,
    parent_source: Optional[Dict[str, Any]],
    use_cloud_function: bool,
    research_topic: str,
    entity_types: List[str],
    cache_manager: CacheManager,
) -> Optional[Work]:
    print(f"Processing link: {url}")

    try:
        work: Work = await get_paper_details(url)
        if work is None:
            abstract = f"Crawled from {url}"
            name = "Unknown crawled link"
            if parent_source is not None:
                if isinstance(parent_source, dict):
                    abstract = f"Crawled from {url} with parent {parent_source['name']} | {parent_source['url']}"
                    name = parent_source["name"]
                else:
                    abstract = f"Crawled from {url} with parent {parent_source.name} | {parent_source.url}"
                    name = parent_source.name

            work = Work(
                id=hashlib.md5(url.encode()).hexdigest(),
                work_type=WORK_TYPES.UNKNOWN,
                name=name,
                authors=[],
                abstract=abstract,
                url=url,
                source_type=source_type,
            )

        if work:
            print(f"Work found: {work}")
            result: Entity = await _download_and_evaluate(
                work,
                depth + 1,
                max_depth,
                use_cloud_function,
                research_topic,
                entity_types,
                cache_manager,
            )
            return result
        else:
            print(f"No work found for {url}")
            return None
    except Exception as e:
        print(f"Error processing link {url}: {e}")
        traceback.print_exc()
        return None



import asyncio
import json
import os
import random
import sys
from pathlib import Path
from typing import List, Set, Optional, Tuple

from bs4 import BeautifulSoup, NavigableString
from playwright.async_api import Page, Browser


def get_proxy():
    print("Getting proxy")
    # if proxies.json doesn't exist, return None
    if not Path("proxies.json").exists():
        return None

    # load proxies.json
    with open("proxies.json") as f:
        proxies = json.load(f)

    # select a random proxy
    proxy = random.choice(proxies)

    # format the proxy for Playwright
    proxy_url = f"{proxy['ip']}:{proxy['port']}"

    return {
        "server": proxy_url,
        # Optionally, if the proxy requires authentication, you can include "username" and "password"
        # "username": proxy.get("username"),
        # "password": proxy.get("password"),
    }


# Lists of keywords and elements to be blacklisted during scraping
link_blacklist: List[str] = [
    "home",
    "next",
    "about us",
    "contact",
    "log in",
    "account",
    "sign",
    "sign up",
    "sign in",
    "sign out",
    "privacy",
    "close",
    "privacy policy",
    "terms of service",
    "terms and conditions",
    "terms",
    "conditions",
    "privacy",
    "legal",
    "guidelines",
    "filter",
    "theme",
    "english",
    "accessibility",
    "authenticate",
    "join",
    "edition",
    "subscribe",
    "news",
    "home",
    "blog",
    "jump to",
    "español",
    "world",
    "europe",
    "politics",
    "profile",
    "election",
    "health",
    "business",
    "tech",
    "sports",
    "advertise",
    "advertising",
    "ad",
    "banner",
    "sponsor",
    "promotion",
    "promoted",
]

element_blacklist: List[str] = [
    "sidebar",
    "nav",
    "footer",
    "header",
    "menu",
    "account",
    "login",
    "form",
    "search",
    "advertisement",
    "masthead",
    "popup",
    "overlay",
    "floater",
    "modal",
    "noscript",
    "iframe",
    "script",
    "style",
    "head",
    "meta",
]


def _check_for_playwright() -> Path:
    """
    Check for Playwright installation and install if not present.

    This function checks if the Playwright browser (Chromium) is installed on the system.
    If not found, it installs it. It then returns the path to the browser executable.

    Returns
    -------
    Path
        The path to the Chromium browser executable.

    Notes
    -----
    This function is platform-aware and handles differences between Windows, macOS, and Linux.
    """
    if sys.platform.startswith("win"):
        browsers_path = Path(os.getenv("LOCALAPPDATA", "")) / "ms-playwright"
    elif sys.platform == "darwin":
        browsers_path = Path.home() / "Library" / "Caches" / "ms-playwright"
    else:  # Linux and other Unix-like OSes
        browsers_path = Path.home() / ".cache" / "ms-playwright"

    # Search for any folder that contains the word 'chromium'
    chromium_folders = list(browsers_path.glob("*chromium*"))

    chromium_installed = bool(chromium_folders)
    if chromium_installed:
        for folder in chromium_folders:
            print(f"Found chromium folder: {folder}")
    else:
        print("No chromium folder found")

    if not chromium_installed:
        print("Browser binaries not found. Installing now...")
        # Run playwright install chromium in subprocess
        import subprocess

        result = subprocess.run(
            ["playwright", "install", "chromium"], capture_output=True, text=True
        )
        print("Installation output:")
        print(result.stdout)
        if result.stderr:
            print("Error output:")
            print(result.stderr)
    else:
        print("Browser binaries are already installed.")

    # Get the browser path from chromium_folders
    chromium_folders = list(browsers_path.glob("*chromium*"))
    browser_path = chromium_folders[0]

    # Find the specific browser executable based on the platform
    if sys.platform == "darwin":
        browser_path = (
            browser_path
            / "chrome-mac"
            / "Chromium.app"
            / "Contents"
            / "MacOS"
            / "Chromium"
        )
    elif sys.platform == "linux":
        browser_path = browser_path / "chrome-linux" / "chrome"
    elif sys.platform == "win32":
        browser_path = browser_path / "chrome-win" / "chrome.exe"

    return browser_path


async def get_browser_and_page(
    playwright, capsolver_api_key: Optional[str] = None
) -> Tuple[Browser, Page]:
    """
    Initialize and return a Playwright browser and page with stealth settings.

    This function sets up a browser with stealth settings to avoid detection,
    and optionally includes a CAPTCHA solver extension.

    Parameters
    ----------
    playwright : Playwright
        The Playwright instance to use for launching the browser.
    capsolver_api_key : Optional[str], optional
        The API key for the CAPTCHA solver service (default is None).

    Returns
    -------
    Tuple[Browser, Page]
        A tuple containing the initialized Browser and Page objects.

    Notes
    -----
    This function uses the playwright_stealth module to apply stealth settings
    and the capsolver_extension_python for CAPTCHA solving capabilities.
    """
    browser_path = _check_for_playwright()

    from playwright_stealth import stealth_async

    # Check if CAPSOLVER_API_KEY exists
    if capsolver_api_key is None:
        capsolver_api_key = os.getenv("CAPSOLVER_API_KEY")

    if capsolver_api_key is None or capsolver_api_key == "":
        print(
            "Warning: No capsolver API key provided. Attempting to launch browser without a captcha solver"
        )

    from capsolver_extension_python import Capsolver

    extension_path = Capsolver(api_key=capsolver_api_key).load(
        with_command_line_option=False
    )
    user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36"
    browser = await playwright.chromium.launch(
        headless=True,
        executable_path=browser_path,
        args=[
            "--disable-extensions-except=" + extension_path,
            "--load-extension=" + extension_path,
            "--user-agent=" + user_agent,
        ],
        proxy=get_proxy(),
    )
    page = await browser.new_page()
    await stealth_async(page)
    return browser, page


async def _extract_plain_text(page: Page) -> str:
    """
    Extract plain text content from a Playwright page.

    This function extracts the text content from the page, removing unwanted elements
    and formatting the text for readability.

    Parameters
    ----------
    page : Page
        The Playwright Page object from which to extract text.

    Returns
    -------
    str
        The extracted and formatted plain text content.

    Notes
    -----
    This function uses BeautifulSoup for HTML parsing and applies custom logic
    to differentiate between inline and block elements for better formatting.
    """
    content = await page.content()
    soup = BeautifulSoup(content, "html.parser")

    # Remove scripts and style elements
    for element in soup(["script", "style"]):
        element.decompose()

    # Define blacklists
    link_blacklist = [
        "unwanted_link_keyword1",
        "unwanted_link_keyword2",
    ]  # Adjust as needed

    # Remove blacklisted links
    for link in soup.find_all("a"):
        href = link.get("href", "").lower()
        text = link.text.lower()
        if any(keyword in href or keyword in text for keyword in link_blacklist):
            link.decompose()

    # Define inline and block elements
    inline_elements: Set[str] = {
        "span",
        "a",
        "button",
        "strong",
        "em",
        "i",
        "b",
        "small",
        "code",
    }
    block_elements: Set[str] = {
        "div",
        "main",
        "section",
        "article",
        "section",
        "p",
        "li",
        "ol",
        "ul",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "blockquote",
    }

    def process_element(element):
        """Recursively process HTML elements to extract formatted text."""
        if isinstance(element, NavigableString):
            return element.strip()

        if element.name in inline_elements:
            return (
                " "
                + " ".join(
                    process_element(child)
                    for child in element.children
                    if child.name != "br"
                ).strip()
                + " "
            )
        elif element.name in block_elements:
            return (
                "\n"
                + "\n".join(
                    process_element(child) for child in element.children
                ).strip()
                + "\n"
            )
        else:
            return " ".join(
                process_element(child) for child in element.children
            ).strip()

    text = process_element(soup.body or soup)  # Start processing from body or root
    text = " ".join(text.split())  # Normalize whitespace
    text = (
        text.replace(" \n ", "\n").replace("\n ", "\n").replace(" \n", "\n").strip()
    )  # Final cleanup

    return text


def _resolve_with_wayback_machine(link: str) -> str:
    """
    Resolve a URL using the Wayback Machine.

    This function attempts to find the most recent archived version of a given URL
    using the Wayback Machine.

    Parameters
    ----------
    link : str
        The URL to resolve.

    Returns
    -------
    str
        The URL of the most recent archived version of the page.

    Notes
    -----
    This function uses the waybackpy library to interact with the Wayback Machine API.
    """
    from waybackpy import Url

    user_agent = "Mozilla/5.0 (Windows NT 5.1; rv:40.0) Gecko/20100101 Firefox/40.0"
    wayback_url = Url(link, user_agent)
    wayback_url.newest()
    url = wayback_url.archive_url
    return url


# List of domains for which we prefer to use Wayback Machine resolution
wayback_resolution_whitelist: List[str] = [
    "nytimes",
    "wsj",
    "nyt",
    "latimes",
    "washingtonpost",
    "bbc",
    "cnn",
    "foxnews",
    "nbcnews",
    "cnbc",
    "reuters",
    "apnews",
    "bloomberg",
    "businessinsider",
    "businesswire",
    "theverge",
]


async def browser_scrape(link: str) -> str:
    """
    Asynchronously scrape a webpage to extract text content.

    This function uses Playwright to navigate to a webpage, render its content,
    and extract the text. It includes fallback mechanisms such as using the
    Wayback Machine for certain domains.

    Parameters
    ----------
    link : str
        The URL of the webpage to scrape.

    Returns
    -------
    str
        The extracted text content from the webpage.

    Notes
    -----
    This function includes error handling and attempts to use the Wayback Machine
    as a fallback for certain whitelisted domains if the initial scrape fails.
    """
    from playwright.async_api import async_playwright

    async with async_playwright() as p:
        try:
            browser, page = await get_browser_and_page(p)
            if any(keyword in link for keyword in wayback_resolution_whitelist):
                new_link = _resolve_with_wayback_machine(link)
                link = new_link if new_link else link

            # Validate the link
            if not link.startswith("http://") and not link.startswith("https://"):
                raise ValueError(f"Invalid link: {link}")
            print(f"Navigating to link: {link}")

            await page.goto(link)
            # Wait for the page to load
            await page.wait_for_load_state("networkidle")
            # Additional wait to handle potential CAPTCHAs
            await asyncio.sleep(2)

            extracted_text = await _extract_plain_text(page)
            return extracted_text
        except Exception as e:
            print(f"Error during scraping: {e}")
            # If we've already used Wayback Machine, don't try again
            if any(keyword in link for keyword in wayback_resolution_whitelist):
                return ""
            # Try to resolve with Wayback Machine
            new_link = _resolve_with_wayback_machine(link)
            if new_link and new_link != link:
                return await browser_scrape(new_link)
            else:
                return ""
        finally:
            await browser.close()



import asyncio
import os
import re
import time
from typing import Optional, Any, Dict
from urllib.parse import urljoin, urlparse

import aiohttp


class ThrottledClientSession(aiohttp.ClientSession):
    """
    A rate-throttled client session class inherited from aiohttp.ClientSession.

    This class implements a leaky bucket algorithm to limit the rate of requests.

    Attributes
    ----------
    rate_limit : float, optional
        The maximum number of requests per second.
    MIN_SLEEP : float
        The minimum sleep time between requests.

    Methods
    -------
    close()
        Close the rate-limiter's "bucket filler" task.
    _filler(rate_limit: float = 1)
        Filler task to implement the leaky bucket algorithm.
    _allow()
        Check if a request is allowed based on the rate limit.
    _request(*args, **kwargs)
        Throttled version of the parent class's _request method.

    Usage
    -----
    Replace `session = aiohttp.ClientSession()`
    with `session = ThrottledClientSession(rate_limit=15)`

    Notes
    -----
    See https://stackoverflow.com/a/60357775/107049 for more details.
    """

    MIN_SLEEP: float = 0.1

    def __init__(
        self, rate_limit: Optional[float] = None, *args: Any, **kwargs: Any
    ) -> None:
        """
        Initialize the ThrottledClientSession.

        Parameters
        ----------
        rate_limit : float, optional
            The maximum number of requests per second.
        *args : Any
            Variable length argument list to be passed to the parent class.
        **kwargs : Any
            Arbitrary keyword arguments to be passed to the parent class.

        Raises
        ------
        ValueError
            If the rate_limit is not positive.
        """
        super().__init__(*args, **kwargs)
        self.rate_limit: Optional[float] = rate_limit
        self._fillerTask: Optional[asyncio.Task] = None
        self._queue: Optional[asyncio.Queue] = None
        self._start_time: float = time.time()
        if rate_limit is not None:
            if rate_limit <= 0:
                raise ValueError("rate_limit must be positive")
            self._queue = asyncio.Queue(min(2, int(rate_limit) + 1))
            self._fillerTask = asyncio.create_task(self._filler(rate_limit))

    def _get_sleep(self) -> Optional[float]:
        """
        Calculate the sleep time between requests based on the rate limit.

        Returns
        -------
        Optional[float]
            The sleep time in seconds, or None if no rate limit is set.
        """
        if self.rate_limit is not None:
            return max(1 / self.rate_limit, self.MIN_SLEEP)
        return None

    async def close(self) -> None:
        """
        Close the rate-limiter's "bucket filler" task.

        This method cancels the filler task and waits for it to complete
        before closing the session.
        """
        if self._fillerTask is not None:
            self._fillerTask.cancel()
        try:
            await asyncio.wait_for(self._fillerTask, timeout=0.5)
        except (asyncio.TimeoutError, asyncio.CancelledError) as err:
            print(str(err))
        await super().close()

    async def _filler(self, rate_limit: float = 1) -> None:
        """
        Filler task to implement the leaky bucket algorithm.

        This method continuously adds tokens to the bucket (queue) based on
        the specified rate limit.

        Parameters
        ----------
        rate_limit : float, optional
            The rate at which to add tokens to the bucket (default is 1).
        """
        try:
            if self._queue is None:
                return
            self.rate_limit = rate_limit
            sleep = self._get_sleep()
            updated_at = time.monotonic()
            fraction = 0
            extra_increment = 0
            for i in range(0, self._queue.maxsize):
                self._queue.put_nowait(i)
            while True:
                if not self._queue.full():
                    now = time.monotonic()
                    increment = rate_limit * (now - updated_at)
                    fraction += increment % 1
                    extra_increment = fraction // 1
                    items_2_add = int(
                        min(
                            self._queue.maxsize - self._queue.qsize(),
                            int(increment) + extra_increment,
                        )
                    )
                    fraction = fraction % 1
                    for i in range(0, items_2_add):
                        self._queue.put_nowait(i)
                    updated_at = now
                await asyncio.sleep(sleep)
        except asyncio.CancelledError:
            pass
        except Exception as err:
            print(str(err))

    async def _allow(self) -> None:
        """
        Check if a request is allowed based on the rate limit.

        This method waits for a token from the queue before allowing a request.
        """
        if self._queue is not None:
            await self._queue.get()
            self._queue.task_done()
        return None

    async def _request(self, *args: Any, **kwargs: Any) -> aiohttp.ClientResponse:
        """
        Throttled version of the parent class's _request method.

        This method ensures that requests are made according to the rate limit.

        Parameters
        ----------
        *args : Any
            Variable length argument list to be passed to the parent's _request method.
        **kwargs : Any
            Arbitrary keyword arguments to be passed to the parent's _request method.

        Returns
        -------
        aiohttp.ClientResponse
            The response from the server.
        """
        await self._allow()
        return await super()._request(*args, **kwargs)


async def cloud_function_request(
    request_type: str,
    params: Dict[str, Any],
    url: Optional[str] = os.getenv("CLOUD_FUNCTION_URL"),
) -> Optional[Dict[str, Any]]:
    """
    Make a request to a cloud function.

    This function sends a POST request to the specified cloud function URL
    with the given request type and parameters.

    Parameters
    ----------
    request_type : str
        The type of request to be made to the cloud function.
    params : Dict[str, Any]
        The parameters to be sent with the request.
    url : Optional[str], optional
        The URL of the cloud function (default is None, which will use the
        CLOUD_FUNCTION_URL environment variable).

    Returns
    -------
    Optional[Dict[str, Any]]
        The JSON response from the cloud function if successful, None otherwise.
    """
    async with aiohttp.ClientSession() as session:
        async with session.post(
            url,
            json={"request_type": request_type, **params},
        ) as response:
            if response.status == 200:
                return await response.json()
            else:
                print(f"Error calling cloud function: {response.status}")
                print(await response.text())
                return None


def validate_url(url: str) -> str:
    """
    Validate and normalize a URL.

    This function checks if the given URL is valid and normalizes it by
    adding the appropriate scheme if missing.

    Parameters
    ----------
    url : str
        The URL to validate and normalize.

    Returns
    -------
    str
        The validated and normalized URL.

    Raises
    ------
    ValueError
        If the URL is invalid and cannot be normalized.
    """
    if url is None:
        raise ValueError("URL cannot be None")

    # Check if the URL already starts with a valid scheme
    if url.startswith(("http://", "https://")):
        return url

    # If it starts with ftp, throw error
    if url.startswith("ftp"):
        raise ValueError(f"FTP URLs are not supported: {url}")

    # If it starts with www, add https://
    if url.startswith("www."):
        url = f"https://{url}"

    # If it starts with anything 2-8 characters and then a ://, throw error
    if re.match(r"^[a-z]{2,8}://", url):
        raise ValueError(f"Invalid URL: {url}")

    # Check for a valid domain structure
    domain_pattern = r"^([a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,}(/.*)?$"
    if re.match(domain_pattern, url, re.IGNORECASE):
        return f"https://{url}"

    # Try to parse the URL
    parsed = urlparse(url)
    if parsed.netloc:
        return f"https://{parsed.netloc}{parsed.path}"

    # If all else fails, raise an exception
    raise ValueError(f"Invalid URL: {url}")


def resolve_url(url: str, base_url: Optional[str] = None) -> str:
    """
    Resolve a URL to its absolute form, handling various edge cases.

    This function takes a URL (which may be relative) and resolves it to
    its absolute form, using a base URL if provided.

    Parameters
    ----------
    url : str
        The URL to resolve.
    base_url : Optional[str], optional
        The base URL to use for relative URLs (default is None).

    Returns
    -------
    str
        The resolved absolute URL.

    Raises
    ------
    ValueError
        If the URL cannot be resolved.
    """
    # Handle protocol-relative URLs
    if url.startswith("//"):
        return f"https:{url}"

    # Handle URLs starting with 'www.'
    if url.startswith("www."):
        url = f"https://{url}"

    # Handle relative URLs
    if (
        url.startswith("/")
        or url.startswith("./")
        or url.startswith("../")
        or not urlparse(url).netloc
    ):
        if not base_url:
            raise ValueError(f"Base URL is required to resolve relative URL: {url}")
        return urljoin(base_url, url)

    # Handle URLs without a scheme
    if not urlparse(url).scheme:
        url = f"https://{url}"

    # Validate the URL
    try:
        parsed = urlparse(url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(f"Invalid URL: {url}")

        # Normalize the URL
        return f"{parsed.scheme}://{parsed.netloc}{parsed.path or '/'}{f'?{parsed.query}' if parsed.query else ''}{f'#{parsed.fragment}' if parsed.fragment else ''}"
    except Exception as e:
        raise ValueError(f"Failed to parse URL {url}: {str(e)}")
